# [[file:../../org/pyretechnics.org::load-landfire-imports][load-landfire-imports]]
import cython
import numpy as np
import rasterio
from rasterio.enums import Resampling
if cython.compiled:
    from cython.cimports.pyretechnics.space_time_cube import SpaceTimeCube
else:
    from pyretechnics.space_time_cube import SpaceTimeCube
# load-landfire-imports ends here
# [[file:../../org/pyretechnics.org::load-raster][load-raster]]
def raster_metadata(raster):
    return {
        "name"      : raster.name,
        "bands"     : raster.count,
        "rows"      : raster.height,
        "cols"      : raster.width,
        "band_types": {i: dtype for i, dtype in zip(raster.indexes, raster.dtypes)},
        "nodata"    : raster.nodata,
        "crs"       : raster.crs,
        "bounds"    : raster.bounds,
        "transform" : raster.transform,
    }


# TODO: rasterio.open can also load chunks of a raster file: https://pypi.org/project/rasterio/
def maybe_resample_resolution(resolution, divisor, resampling_policy):
    if (resolution % divisor == 0):
        return resolution
    elif resampling_policy == "always_upsample":
        return divisor * (resolution // divisor + 1)
    elif resampling_policy == "nearest_match":
        return divisor * max(1, round(resolution / divisor))
    else:
        raise ValueError("The resampling_policy must be either 'always_upsample' or 'nearest_match'.")


def load_raster(file_path, dtype=None, cube_shape_divisors=(1,1,1),
                resampling_policy="nearest_match",
                resampling_method=Resampling.nearest):
    """
    resampling_policy: "always_upsample" or "nearest_match"
    resampling_method: any rasterio.enums.Resampling method
    """
    with rasterio.open(file_path, "r") as input_raster:
        metadata  = raster_metadata(input_raster)
        bands     = metadata["bands"]
        rows      = metadata["rows"]
        cols      = metadata["cols"]
        (b, r, c) = cube_shape_divisors
        new_bands = maybe_resample_resolution(bands, b, resampling_policy)
        new_rows  = maybe_resample_resolution(rows, r, resampling_policy)
        new_cols  = maybe_resample_resolution(cols, c, resampling_policy)
        if new_bands == bands and new_rows == rows and new_cols == cols:
            return {
                "array"   : input_raster.read(out_dtype=dtype),
                "metadata": metadata,
            }
        else:
            metadata["bands"]     = new_bands
            metadata["rows"]      = new_rows
            metadata["cols"]      = new_cols
            metadata["transform"] = (input_raster.transform
                                     * input_raster.transform.scale(
                                         cols / new_cols,
                                         rows / new_rows,
                                     ))
            array = input_raster.read(
                out_dtype=dtype,
                out_shape=(new_bands, new_rows, new_cols),
                resampling=resampling_method,
            )
            return {
                "array"   : array,
                "metadata": metadata,
            }
# load-raster ends here
# [[file:../../org/pyretechnics.org::load-and-convert-landfire-rasters][load-and-convert-landfire-rasters]]
landfire_array_conversions = {
    #==============================================================================
    # Layer Name         : (New dtype, Mult), # New Units                [Min-Max]
    #==============================================================================
    "elevation"          : ("float32", 1.00), # meters above sea level   [0-8850]
    "slope"              : ("float32", 0.01), # meters rise / meters run [0-4]
    "aspect"             : ("int16"  , 1   ), # degrees CW from North    [0-359]
    "fuel_model"         : ("int16"  , 1   ), # integer classes          [1-204]
    "canopy_cover"       : ("float32", 0.01), # canopy area / total area [0-0.95]
    "canopy_height"      : ("float32", 0.10), # meters                   [0-51]
    "canopy_base_height" : ("float32", 0.10), # meters                   [0-10]
    "canopy_bulk_density": ("float32", 0.01), # kilograms/meters^3       [0-0.45]
}


def load_and_convert_landfire_rasters(landfire_file_paths):
    landfire_rasters = {}

    for name, path in landfire_file_paths.items():
        (dtype, multiplier) = landfire_array_conversions[name]
        landfire_rasters[name] = load_raster(path, dtype)
        if multiplier != 1:
            array  = landfire_rasters[name]["array"]
            nodata = landfire_rasters[name]["metadata"]["nodata"]
            array[array != nodata] *= multiplier

    return landfire_rasters
# load-and-convert-landfire-rasters ends here
# [[file:../../org/pyretechnics.org::verify-raster-constraints][verify-raster-constraints]]
def verify_cube_compatible_dimensions(cube_shape, rasters):
    cube_shape_ = np.asarray(cube_shape)
    for r in rasters:
        raster_shape = np.asarray((r["metadata"]["bands"],
                                   r["metadata"]["rows"],
                                   r["metadata"]["cols"]))
        if np.any(raster_shape > cube_shape_):
            raise ValueError("Some raster dimensions exceed the space-time cube dimensions.")

    return True


def verify_same_georeferences(rasters):
    georeferences = [
        (r["metadata"]["crs"],
         r["metadata"]["bounds"],
         r["metadata"]["transform"])
        for r in rasters
    ]
    if len(set(georeferences)) == 1:
        return True
    else:
        raise ValueError("All rasters do not share the same georeferences.")


def verify_raster_constraints(cube_shape, rasters):
    return verify_cube_compatible_dimensions(cube_shape, rasters) and verify_same_georeferences(rasters)
# verify-raster-constraints ends here
# [[file:../../org/pyretechnics.org::convert-rasters-to-space-time-cubes][convert-rasters-to-space-time-cubes]]
def convert_rasters_to_space_time_cubes(cube_shape, raster_dict):
    fn_dict = {}

    for name, raster in raster_dict.items():
        fn_dict[name] = SpaceTimeCube(cube_shape, raster["array"])

    return fn_dict
# convert-rasters-to-space-time-cubes ends here
# [[file:../../org/pyretechnics.org::read-landfire-rasters-as-space-time-cubes][read-landfire-rasters-as-space-time-cubes]]
def read_landfire_rasters_as_space_time_cubes(cube_shape, landfire_file_paths):
    landfire_rasters = load_and_convert_landfire_rasters(landfire_file_paths)
    if verify_raster_constraints(cube_shape, landfire_rasters.values()):
        return convert_rasters_to_space_time_cubes(cube_shape, landfire_rasters)
# read-landfire-rasters-as-space-time-cubes ends here



import cython
import cython as cy
import numpy as np
from numpy.random import Generator
if cython.compiled:
    from cython.cimports.pyretechnics.cy_types import pyidx
else:
    from pyretechnics.py_types import pyidx


@cy.cclass
class BufferedRandGen:
    numpy_rand     : Generator
    uniform_buf    : cy.float[::1]
    uniform_pos    : pyidx
    normal_buf     : cy.float[::1]
    normal_pos     : pyidx
    poisson16_buf  : cy.longlong[::1] # An array of Poisson(16) draws.
    poisson16_pos  : pyidx
    poisson1_buf   : cy.longlong[::1] # An array of Poisson(1) draws.
    poisson1_pos   : pyidx
    poisson_exp_buf: cy.double[::1]   # An array of Exponential(1) draws, with the first value potentially modified.
    poisson_exp_pos: pyidx
    # NOTE: INVARIANT poisson_exp_buf[poisson_exp_pos] is the distance to the next Poisson Process point.


    def __cinit__(self, numpy_rand):
        self.numpy_rand = numpy_rand
        __reset_uniform_buffer(self)
        __reset_normal_buffer(self)
        self.poisson16_buf = self.numpy_rand.poisson(lam=16.0, size=1024).astype(np.int64)
        self.poisson1_buf  = self.numpy_rand.poisson(lam=1.0, size=1024).astype(np.int64)
        __reset_exp_buffer(self)


    @cy.cfunc
    @cy.exceptval(check=False)
    def next_poisson(self, M: cy.double) -> cy.longlong:
        """
        A method for efficiently drawing from a Poisson distribution of (very) small mean,
        doing fewer random draws than method calls.

        Returns a random draw from a Poisson distribution of mean M.
        """
        if M > 0.0:
            ret: cy.longlong = 0
            # Theorem: A sum of independent Poisson variables is Poisson-distributed,
            #          and its mean is the sum of the means.
            # Draw repeatedly from Poisson(16)
            while M >= 16.0:
                ret += __next_poisson16(self)
                M   -= 16.0
            # Draw repeatedly from Poisson(1)
            while M >= 1.0:
                ret += __next_poisson1(self)
                M   -= 1.0
            # Now we draw efficiently from a Poisson distribution of fractional mean.
            # This relies on using a sequence of arrival times distributed i.i.d. as Exponential(1).
            ret_frac: cy.longlong = 0
            pos     : pyidx       = self.poisson_exp_pos
            next_Sk : cy.double   = self.poisson_exp_buf[pos]
            while M >= next_Sk: # This will be rare.
                ret_frac += 1
                M        -= next_Sk
                pos      += 1
                if pos >= 64:
                    __reset_exp_buffer(self)
                    pos = self.poisson_exp_pos
                next_Sk = self.poisson_exp_buf[pos]
            self.poisson_exp_buf[pos] = next_Sk - M
            if ret_frac > 0: # We moved the cursor.
                self.poisson_exp_pos = pos
            ret += ret_frac
            return ret
        else:
            return 0


    @cy.cfunc
    @cy.exceptval(check=False)
    def next_uniform(self) -> cy.float:
        if not(self.uniform_pos < 1024):
            __reset_uniform_buffer(self)
        ret: cy.float     = self.uniform_buf[self.uniform_pos]
        self.uniform_pos += 1
        return ret


    @cy.cfunc
    @cy.exceptval(check=False)
    def next_normal(self) -> cy.float:
        if not(self.normal_pos < 1024):
            __reset_normal_buffer(self)
        ret: cy.float    = self.normal_buf[self.normal_pos]
        self.normal_pos += 1
        return ret


@cy.cfunc
@cy.exceptval(check=False)
def __reset_exp_buffer(self: BufferedRandGen) -> cy.void:
    self.poisson_exp_buf = self.numpy_rand.exponential(size=64)
    self.poisson_exp_pos = 0


@cy.cfunc
@cy.exceptval(check=False)
def __reset_uniform_buffer(self: BufferedRandGen) -> cy.void:
    self.uniform_buf = self.numpy_rand.uniform(size=1024).astype(np.float32)
    self.uniform_pos = 0


@cy.cfunc
@cy.exceptval(check=False)
def __reset_normal_buffer(self: BufferedRandGen) -> cy.void:
    self.normal_buf = self.numpy_rand.normal(size=1024).astype(np.float32)
    self.normal_pos = 0


@cy.cfunc
@cy.exceptval(check=False)
def __next_poisson1(self: BufferedRandGen) -> cy.longlong:
    if not(self.poisson1_pos < 1024):
        self.poisson1_buf = self.numpy_rand.poisson(lam=1.0, size=1024).astype(np.int64)
        self.poisson1_pos = 0
    ret: cy.longlong   = self.poisson1_buf[self.poisson1_pos]
    self.poisson1_pos += 1
    return ret


@cy.cfunc
@cy.exceptval(check=False)
def __next_poisson16(self: BufferedRandGen) -> cy.longlong:
    if not(self.poisson16_pos < 1024):
        self.poisson16_buf = self.numpy_rand.poisson(lam=16.0, size=1024).astype(np.int64)
        self.poisson16_pos = 0
    ret: cy.longlong    = self.poisson16_buf[self.poisson16_pos]
    self.poisson16_pos += 1
    return ret



# [[file:../../org/pyretechnics.org::spot-fire-imports][spot-fire-imports]]
import cython
import cython as cy
if cython.compiled:
    from cython.cimports.libc.math import round, sqrt, pow, log, exp, sin, cos
    from cython.cimports.pyretechnics.cy_types import \
        pyidx, vec_xy, coord_yx, coord_tyx, SpreadBehavior, SpotConfig, JumpDistribution
    from cython.cimports.pyretechnics.random import BufferedRandGen
    from cython.cimports.pyretechnics.space_time_cube import ISpaceTimeCube
    import cython.cimports.pyretechnics.conversion as conv
    import cython.cimports.pyretechnics.fuel_models as fm
    import cython.cimports.pyretechnics.surface_fire as sf
else:
    from math import sqrt, pow, log, exp, sin, cos
    from pyretechnics.py_types import \
        pyidx, vec_xy, coord_yx, coord_tyx, SpreadBehavior, SpotConfig, JumpDistribution
    from pyretechnics.random import BufferedRandGen
    from pyretechnics.space_time_cube import ISpaceTimeCube
    import pyretechnics.conversion as conv
    import pyretechnics.fuel_models as fm
    import pyretechnics.surface_fire as sf
# spot-fire-imports ends here
# [[file:../../org/pyretechnics.org::expected-firebrand-production][expected-firebrand-production]]
@cy.cfunc
@cy.exceptval(check=False)
def expected_firebrand_production(fire_behavior           : SpreadBehavior,
                                  elevation_gradient      : vec_xy,
                                  cell_horizontal_area    : cy.float,
                                  firebrands_per_unit_heat: cy.float=1e-6) -> cy.float:
    """
    Return the expected number of firebrands produced by an entire cell when it burns given:
    - fire_behavior            :: a SpreadBehavior struct of surface or crown fire behavior values
      - dphi_dt                   :: phi/min
      - fire_type                 :: 0 (unburned), 1 (surface), 2 (passive_crown), or 3 (active_crown)
      - spread_rate               :: m/min
      - spread_direction          :: (x, y, z) unit vector on the slope-tangential plane
      - fireline_intensity        :: kW/m
      - flame_length              :: m
    - elevation_gradient       :: tuple with these fields
      - dz_dx                     :: rise/run
      - dz_dy                     :: rise/run
    - cell_horizontal_area     :: m^2
    - firebrands_per_unit_heat :: firebrands/kJ
    """
    if fire_behavior.spread_rate == 0.0:
        return 0.0
    else:
        #================================================================================================
        # Calculate the heat output per unit area
        #================================================================================================

        spread_rate         : cy.float = fire_behavior.spread_rate                                  # m/min
        fireline_intensity  : cy.float = fire_behavior.fireline_intensity                           # kW/m
        heat_output_per_area: cy.float = sf.calc_areal_heat_output(spread_rate, fireline_intensity) # kJ/m^2

        #================================================================================================
        # Calculate the slope-adjusted cell area
        #================================================================================================

        (dz_dx, dz_dy)         = elevation_gradient                            # (rise/run, rise/run)
        slope_factor: cy.float = sqrt(1.0 + (dz_dx * dz_dx) + (dz_dy * dz_dy)) # unitless
        cell_area   : cy.float = cell_horizontal_area * slope_factor           # m^2

        #================================================================================================
        # Calculate the expected number of firebrands produced in this cell
        #================================================================================================

        cell_heat_output: cy.float = heat_output_per_area * cell_area            # kJ
        firebrand_count : cy.float = firebrands_per_unit_heat * cell_heat_output # number of firebrands
        return firebrand_count
# expected-firebrand-production ends here
# [[file:../../org/pyretechnics.org::convert-deltas][convert-deltas]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def delta_to_grid_dx(cos_wdir: cy.float, sin_wdir: cy.float, delta_x: cy.float, delta_y: cy.float) -> cy.float:
    """
    Computes the grid-aligned x coordinate of the delta vector, given the wind-aligned [ΔX ΔY] coordinates.
    Returns a signed distance (same unit as ΔX and ΔY).

    NOTE:
    - sin_wdir = wdir_x
    - cos_wdir = wdir_perp_x
    """
    return delta_x * sin_wdir + delta_y * cos_wdir


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def delta_to_grid_dy(cos_wdir: cy.float, sin_wdir: cy.float, delta_x: cy.float, delta_y: cy.float) -> cy.float:
    """
    Computes the grid-aligned y coordinate of the delta vector, given the wind-aligned [ΔX ΔY] coordinates.
    Returns a signed distance (same unit as ΔX and ΔY).

    NOTE:
    - cos_wdir = wdir_y
    - sin_wdir = wdir_perp_y
    """
    return delta_x * cos_wdir + delta_y * sin_wdir


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def distance_to_n_cells(distance: cy.float, cell_size: cy.float) -> cy.int:
    """
    Converts a delta expressed as a signed distance to one expressed as a number of grid cells.
    """
    return int(round(distance / cell_size))
# convert-deltas ends here
# [[file:../../org/pyretechnics.org::resolve-spotting-lognormal-elmfire][resolve-spotting-lognormal-elmfire]]
@cy.cfunc
@cy.exceptval(check=False)
def resolve_exp_delta_x(spot_config: SpotConfig, fireline_intensity: cy.float, wind_speed_20ft: cy.float) -> cy.float:
    """
    Computes the expected value E[ΔX] (in meters) of the downwind spotting distance ΔX given:
    - spot_config        :: a SpotConfig struct of spotting parameters
      - random_seed                  :: seed for a numpy.random.Generator object
      - firebrands_per_unit_heat     :: firebrands/kJ
      - downwind_distance_mean       :: meters
      - fireline_intensity_exponent  :: downwind_distance_mean multiplier [I^fireline_intensity_exponent]
      - wind_speed_exponent          :: downwind_distance_mean multiplier [U^wind_speed_exponent]
      - downwind_variance_mean_ratio :: meters^2 / meter [downwind_variance_mean_ratio = Var(X) / E(X)]
      - crosswind_distance_stdev     :: meters
      - decay_distance               :: meters
    - fireline_intensity :: kW/m
    - wind_speed_20ft    :: m/s
    """
    downwind_distance_mean     : cy.float = spot_config.downwind_distance_mean
    fireline_intensity_exponent: cy.float = spot_config.fireline_intensity_exponent
    wind_speed_exponent        : cy.float = spot_config.wind_speed_exponent
    return (downwind_distance_mean
            * pow(fireline_intensity, fireline_intensity_exponent)
            * pow(wind_speed_20ft, wind_speed_exponent))


@cy.cfunc
@cy.exceptval(check=False)
def resolve_var_delta_x(spot_config: SpotConfig, exp_delta_x: cy.float) -> cy.float:
    """
    Computes the variance Var[ΔX] (in m^2) of the downwind spotting distance ΔX given:
    - spot_config :: a SpotConfig struct of spotting parameters
      - random_seed                  :: seed for a numpy.random.Generator object
      - firebrands_per_unit_heat     :: firebrands/kJ
      - downwind_distance_mean       :: meters
      - fireline_intensity_exponent  :: downwind_distance_mean multiplier [I^fireline_intensity_exponent]
      - wind_speed_exponent          :: downwind_distance_mean multiplier [U^wind_speed_exponent]
      - downwind_variance_mean_ratio :: meters^2 / meter [downwind_variance_mean_ratio = Var(X) / E(X)]
      - crosswind_distance_stdev     :: meters
      - decay_distance               :: meters
    - exp_delta_x :: meters (E[ΔX])
    """
    downwind_variance_mean_ratio: cy.float = spot_config.downwind_variance_mean_ratio
    return downwind_variance_mean_ratio * exp_delta_x


@cy.cfunc
@cy.exceptval(check=False)
def lognormal_mu_from_moments(mean: cy.float, variance: cy.float) -> cy.float:
    """
    TODO: Add docstring
    """
    m2: cy.float = mean * mean
    return log(m2 / sqrt(m2 + variance))


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def lognormal_sigma_from_moments(mean: cy.float, variance: cy.float) -> cy.float:
    """
    TODO: Add docstring
    """
    return sqrt(log(1.0 + variance / (mean * mean)))


@cy.cfunc
@cy.exceptval(check=False)
def resolve_lognormal_params(spot_config       : SpotConfig,
                             fireline_intensity: cy.float,
                             wind_speed_20ft   : cy.float) -> tuple[cy.float, cy.float]:
    """
    TODO: Add docstring
    """
    exp_delta_x         : cy.float = resolve_exp_delta_x(spot_config, fireline_intensity, wind_speed_20ft)
    var_delta_x         : cy.float = resolve_var_delta_x(spot_config, exp_delta_x)
    prob_lognormal_mu   : cy.float = lognormal_mu_from_moments(exp_delta_x, var_delta_x)
    prob_lognormal_sigma: cy.float = lognormal_sigma_from_moments(exp_delta_x, var_delta_x)
    return (prob_lognormal_mu, prob_lognormal_sigma)
# resolve-spotting-lognormal-elmfire ends here
# [[file:../../org/pyretechnics.org::resolve-spotting-normal-elmfire][resolve-spotting-normal-elmfire]]
# When will we have the default sigma_Y > E[ΔX]?
# It can be seen that this nonsensical situation
# happens iff sigma_X exceeds the following number:
#
# sqrt(log(1.0 + (0.88 ** 2.0) / (0.92 * 0.47))
#
# => 1.0131023746492023
sigma_y_scalar_m = cy.declare(cy.double, 0.92 * 0.47 / (0.88 * 0.88))


@cy.cfunc
@cy.exceptval(check=False)
def himoto_resolve_default_sigma_y_from_lognormal_params(mu_x: cy.float, sigma_x: cy.float) -> cy.float:
    es2h      : cy.float = exp((sigma_x * sigma_x) / 2.0)
    avg_deltax: cy.float = exp(mu_x) * es2h
    return sigma_y_scalar_m * avg_deltax * (es2h + 1.0) * (es2h - 1.0) # meters


@cy.cfunc
@cy.exceptval(check=False)
def himoto_resolve_default_sigma_y(spot_config       : SpotConfig,
                                   fireline_intensity: cy.float,
                                   wind_speed_20ft   : cy.float) -> cy.float:
    ln_params: tuple[cy.float, cy.float] = resolve_lognormal_params(spot_config, fireline_intensity, wind_speed_20ft)
    mu_x     : cy.float                  = ln_params[0] # dimensionless (log-space)
    sigma_x  : cy.float                  = ln_params[1] # dimensionless (log-space)
    return himoto_resolve_default_sigma_y_from_lognormal_params(mu_x, sigma_x) # meters


@cy.cfunc
@cy.exceptval(check=False)
def resolve_crosswind_distance_stdev(spot_config       : SpotConfig,
                                     fireline_intensity: cy.float,
                                     wind_speed_20ft   : cy.float) -> cy.float:
    crosswind_distance_stdev: cy.float = spot_config.crosswind_distance_stdev
    if crosswind_distance_stdev != 0.0:
        return crosswind_distance_stdev # meters
    else:
        return himoto_resolve_default_sigma_y(spot_config, fireline_intensity, wind_speed_20ft) # meters
# resolve-spotting-normal-elmfire ends here
# [[file:../../org/pyretechnics.org::sardoy-firebrand-dispersal][sardoy-firebrand-dispersal]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def sample_normal(rng: BufferedRandGen, mu: cy.float, sd: cy.float) -> cy.float:
    """
    Returns sample from normal/gaussian distribution given mu and sd.
    """
    return mu + sd * rng.next_normal()


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def sample_lognormal(rng: BufferedRandGen, mu: cy.float, sd: cy.float) -> cy.float:
    """
    Returns sample from log-normal distribution given mu and sd.
    """
    return exp(sample_normal(rng, mu, sd))


@cy.cfunc
@cy.exceptval(check=False)
def resolve_JumpDistribution(spot_config       : SpotConfig,
                             fireline_intensity: cy.float,
                             wind_speed_20ft   : cy.float) -> JumpDistribution:
    ln_params: tuple[cy.float, cy.float] = resolve_lognormal_params(spot_config, fireline_intensity, wind_speed_20ft)
    # Initialize a new JumpDistribution
    return JumpDistribution(
        mu_x    = ln_params[0], # dimensionless (log-space)
        sigma_x = ln_params[1], # dimensionless (log-space)
        sigma_y = resolve_crosswind_distance_stdev(spot_config, fireline_intensity, wind_speed_20ft), # meters
    )


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def sample_downwind_jump(jd: JumpDistribution, random_generator: BufferedRandGen) -> cy.float:
    return sample_lognormal(random_generator, jd.mu_x, jd.sigma_x)


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def sample_crosswind_jump(jd: JumpDistribution, random_generator: BufferedRandGen) -> cy.float:
    return sample_normal(random_generator, 0.0, jd.sigma_y)
# sardoy-firebrand-dispersal ends here
# [[file:../../org/pyretechnics.org::schroeder-ignition-probability][schroeder-ignition-probability]]
@cy.cfunc
@cy.exceptval(check=False)
def heat_of_preignition(temperature: cy.float, fine_fuel_moisture: cy.float) -> cy.float:
    """
    Returns heat of preignition given:
    - temperature        :: degrees Celsius
    - fine_fuel_moisture :: 0-1

    Q_ig = 144.512 - 0.266 * T_o - 0.00058 * (T_o)^2 - T_o * M + 18.54 * (1 - exp(-15.1 * M)) + 640 * M (eq. 10)
    """
    T_o: cy.float = temperature
    M  : cy.float = fine_fuel_moisture
    # Heat required to reach ignition temperature
    Q_a: cy.float = 144.512 - 0.266 * T_o - 0.00058 * (T_o * T_o)
    # Heat required to raise moisture to the boiling point
    Q_b: cy.float = -T_o * M
    # Heat of desorption
    Q_c: cy.float = 18.54 * (1.0 - exp(-15.1 * M))
    # Heat required to vaporize moisture
    Q_d: cy.float = 640.0 * M
    return Q_a + Q_b + Q_c + Q_d


@cy.ccall
@cy.exceptval(check=False)
def schroeder_ignition_probability(temperature: cy.float, fine_fuel_moisture: cy.float) -> cy.float:
    """
    Returns the probability of spot fire ignition (Schroeder 1969) given:
    - temperature        :: degrees Celsius
    - fine_fuel_moisture :: 0-1

    X    = (400 - Q_ig) / 10
    P(I) = (0.000048 * X^4.3) / 50 (pg. 15)
    """
    Q_ig: cy.float = heat_of_preignition(temperature, fine_fuel_moisture)
    X   : cy.float = max(0.0, 400.0 - Q_ig) * 0.1
    P_I : cy.float = 0.000048 * pow(X, 4.3) * 0.02
    return min(P_I, 1.0)
# schroeder-ignition-probability ends here
# [[file:../../org/pyretechnics.org::firebrand-flight-survival-probability][firebrand-flight-survival-probability]]
@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def firebrand_flight_survival_probability(spotting_distance: cy.float, decay_distance: cy.float) -> cy.float:
    """
    Returns the probability that a firebrand will survive its flight (Perryman 2012) given:
    - spotting_distance :: meters (d)
    - decay_distance    :: meters (1/lambda)

    P(Survival) = exp(-d * lambda)
    """
    return exp(-spotting_distance / decay_distance)
# firebrand-flight-survival-probability ends here
# [[file:../../org/pyretechnics.org::firebrands-time-of-ignition][firebrands-time-of-ignition]]
# FIXME: unused
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def albini_firebrand_maximum_height(firebrand_diameter: cy.float) -> cy.float:
    return 0.39e5 * firebrand_diameter


@cy.cfunc
@cy.exceptval(check=False)
def albini_t_max(flame_length: cy.float) -> cy.float:
    """
    Returns the time of spot ignition using Albini1979spot in minutes given:
    - flame_length :: meters [z_F]

    a           = 5.963                                                             (D33)
    b           = a - 1.4                                                           (D34)
    D           = 0.003
    z           = 0.39 * D * 10^5
    w_F         = 2.3 * z_F^0.5                                                     (A58)
    t_c         = 1
    t_o         = t_c / (2 * z_F / w_F)
    travel_time = t_1 + t_2 + t_3 = 1.2 + (a / 3) * (((b + (z/z_F)) / a)^3/2 - 1)   (D43)
    """
    a        : cy.float = 5.963  # dimensionless constant from (D33)
    b        : cy.float = 4.563  # dimensionless constant from (D34)
    z        : cy.float = 117.0  # maximum altitude of firebrands in meters [derived for (D44) in (Albini1979spot)]
    z_F      : cy.float = flame_length                      # m
    w_F      : cy.float = 2.3 * sqrt(flame_length)          # m/s
    charact_t: cy.float = conv.sec_to_min(2.0 * z_F / w_F)  # min
    # The following dimensionless factor is equal to t_T - t_o, with t_T defined by (D43) in Albini1979spot.
    u          : cy.float = (b + z / z_F) / a
    u3_2       : cy.float = u * sqrt(u) # Faster than ** 1.5
    travel_time: cy.float = 1.2 + (a / 3.0) * (u3_2 - 1.0)
    return charact_t * travel_time


# FIXME: Consider removing t_max from this calculation for performance.
@cy.cfunc
@cy.exceptval(check=False)
def spot_ignition_time(time_of_arrival: cy.float, flame_length: cy.float) -> cy.float:
    """
    Returns the time of spot ignition using Albini 1979 and Perryman 2012 in minutes given:
    - time_of_arrival :: minutes
    - flame_length    :: meters

    t_spot = time_of_arrival + (2 * t_max) + t_ss
    """
    t_max         : cy.float = albini_t_max(flame_length)
    t_steady_state: cy.float = 20.0 # period of building up to steady state from ignition (min)
    return time_of_arrival + 2.0 * t_max + t_steady_state
# firebrands-time-of-ignition ends here
# [[file:../../org/pyretechnics.org::spread-firebrands][spread-firebrands]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def is_in_bounds(y: pyidx, x: pyidx, rows: pyidx, cols: pyidx) -> cy.bint:
    """
    Returns True if the grid coordinate (y,x) lies within the bounds [0,rows) by [0,cols).
    """
    return (y >= 0) and (x >= 0) and (y < rows) and (x < cols)


# FIXME: Memoize this function with a 2D (y,x) array cache
@cy.cfunc
@cy.exceptval(check=False)
def is_burnable_cell(fuel_model_cube: ISpaceTimeCube, t: pyidx, y: pyidx, x: pyidx) -> cy.bint:
    """
    Returns True if the space-time coordinate (t,y,x) contains a burnable fuel model.
    """
    fuel_model_number: cy.int = cy.cast(cy.int, fuel_model_cube.get(t,y,x))
    return fm.is_burnable_fuel_model_number(fuel_model_number)


@cy.cfunc
@cy.exceptval(check=False)
def cast_firebrand(rng                        : BufferedRandGen,
                   fuel_model_cube            : ISpaceTimeCube,
                   temperature_cube           : ISpaceTimeCube,
                   fuel_moisture_dead_1hr_cube: ISpaceTimeCube,
                   fire_type_matrix           : cy.uchar[:,::1],
                   rows                       : pyidx,
                   cols                       : pyidx,
                   cell_height                : cy.float,
                   cell_width                 : cy.float,
                   source_t                   : pyidx,
                   source_y                   : pyidx,
                   source_x                   : pyidx,
                   decay_distance             : cy.float,
                   cos_wdir                   : cy.float,
                   sin_wdir                   : cy.float,
                   jd                         : JumpDistribution) -> coord_yx:
    """
    Draws a random [ΔX, ΔY] pair of signed distances (in meters) from
    the supplied cell, representing the coordinates of the spotting jump in the directions
    parallel and perpendicular to the wind. ΔX will typically be positive (downwind),
    and positive ΔY means to the right of the downwind direction.

    NOTE: If the random draw yields no spot ignition, the source cell coordinates will be returned.
          Calling code should check for that.
    """
    #=======================================================================================
    # Determine where the firebrand will land
    #=======================================================================================

    delta_y: cy.float = sample_crosswind_jump(jd, rng)                         # meters
    delta_x: cy.float = sample_downwind_jump(jd, rng)                          # meters
    grid_dy: cy.float = delta_to_grid_dy(cos_wdir, sin_wdir, delta_x, delta_y) # meters
    grid_dx: cy.float = delta_to_grid_dx(cos_wdir, sin_wdir, delta_x, delta_y) # meters
    # NOTE: It would cause a bug to type the following as pyidx.
    target_y: cy.int = source_y + distance_to_n_cells(grid_dy, cell_height)
    target_x: cy.int = source_x + distance_to_n_cells(grid_dx, cell_width)

    #=======================================================================================
    # Determine whether the firebrand will start a fire or fizzle out
    #=======================================================================================

    if is_in_bounds(target_y, target_x, rows, cols) and fire_type_matrix[target_y,target_x] == 0:
        # Firebrand landed on the grid in an unburned cell
        # Calculate the probability that the firebrand survived its flight and landed while still burning
        spotting_distance          : cy.float = sqrt(grid_dx * grid_dx + grid_dy * grid_dy) # meters
        flight_survival_probability: cy.float = firebrand_flight_survival_probability(spotting_distance,
                                                                                      decay_distance)

        # Roll the dice
        uniform_sample: cy.float = rng.next_uniform()

        if (uniform_sample <= flight_survival_probability
            and is_burnable_cell(fuel_model_cube, source_t, target_y, target_x)):
            # Firebrand landed in a cell with a burnable fuel model, so calculate its ignition probability
            temperature         : cy.float = temperature_cube.get(source_t, target_y, target_x) # degrees Celsius
            fine_fuel_moisture  : cy.float = fuel_moisture_dead_1hr_cube.get(source_t, target_y, target_x) # %
            ignition_probability: cy.float = schroeder_ignition_probability(temperature, fine_fuel_moisture)

            if uniform_sample <= flight_survival_probability * ignition_probability:
                # Firebrand ignited the target cell, so return its coordinates for later processing
                return (target_y, target_x)
    # This code is only reached if the spotting ignition fails.
    # For efficiency, the source cell is used as a sentinel value for a failed spot ignition.
    return (source_y, source_x)


@cy.cfunc
def spread_firebrands(fuel_model_cube            : ISpaceTimeCube,
                      temperature_cube           : ISpaceTimeCube,
                      fuel_moisture_dead_1hr_cube: ISpaceTimeCube,
                      fire_type_matrix           : cy.uchar[:,::1],
                      sim_area_bounds            : coord_yx,
                      cell_height                : cy.float,
                      cell_width                 : cy.float,
                      space_time_coordinate      : coord_tyx,
                      wind_speed_10m             : cy.float, # km/hr (for shame!)
                      upwind_direction           : cy.float, # degrees
                      fireline_intensity         : cy.float,
                      flame_length               : cy.float,
                      time_of_arrival            : cy.float,
                      random_generator           : BufferedRandGen,
                      num_firebrands             : cy.longlong,
                      spot_config                : SpotConfig) -> tuple: # tuple[float, set]|None
    """
    Given these inputs:
    - fuel_model_cube             :: (Lazy)SpaceTimeCube (integer index in fm.fuel_model_table)
    - temperature_cube            :: (Lazy)SpaceTimeCube (degrees Celsius)
    - fuel_moisture_dead_1hr_cube :: (Lazy)SpaceTimeCube (kg moisture/kg ovendry weight)
    - fire_type_matrix            :: 2D byte array (0=unburned, 1=surface, 2=passive_crown, 3=active_crown)
    - sim_area_bounds             :: tuple with these fields
      - rows                         :: number of rows on the simulation grid
      - cols                         :: number of columns on the simulation grid
    - cell_height                 :: meters
    - cell_width                  :: meters
    - space_time_coordinate       :: (t,y,x) coordinate in which the source cell burns
    - upwind_direction            :: degrees clockwise from North
    - wind_speed_10m              :: km/hr
    - fireline_intensity          :: kW/m
    - flame_length                :: m
    - time_of_arrival             :: min
    - random_generator            :: BufferedRandGen
    - num_firebrands              :: number of firebrands to emit from the space_time_coordinate
    - spot_config                 :: SpotConfig struct of spotting parameters
      - downwind_distance_mean       :: meters
      - fireline_intensity_exponent  :: downwind_distance_mean multiplier [I^fireline_intensity_exponent]
      - wind_speed_exponent          :: downwind_distance_mean multiplier [U^wind_speed_exponent]
      - downwind_variance_mean_ratio :: meters^2 / meter [downwind_variance_mean_ratio = Var(X) / E(X)]
      - crosswind_distance_stdev     :: meters
      - decay_distance               :: meters

    casts num_firebrands from the space_time_coordinate into grid cells in the space-time cube, filters out
    all of the firebrands that fizzle out in either burnable or non-burnable fuels, and returns any that ignite
    new spot fires in a tuple with these fields:

    - ignition_time :: minutes
    - ignited_cells :: set of (y,x) grid coordinates
    """
    #=======================================================================================
    # Ensure that the source cell is casting firebrands
    #=======================================================================================

    if num_firebrands > 0:

        #=======================================================================================
        # Ensure that there is wind to transport the firebrands
        #=======================================================================================

        if wind_speed_10m > 0.0:

            #=======================================================================================
            # Unpack all firebrand-related features of the source cell
            #=======================================================================================

            # TODO: OPTIM Get rid of trigonometry here if possible by having callers pass vectors.
            (rows, cols)                          = sim_area_bounds
            (t, y, x)                             = space_time_coordinate
            decay_distance     : cy.float         = spot_config.decay_distance
            downwind_direction : cy.float         = conv.deg_to_rad(conv.opposite_direction(upwind_direction))
            cos_wdir           : cy.float         = cos(downwind_direction)
            sin_wdir           : cy.float         = sin(downwind_direction)
            wind_speed_20ft    : cy.float         = conv.wind_speed_10m_to_wind_speed_20ft(wind_speed_10m) # km/hr
            wind_speed_20ft_mps: cy.float         = conv.km_hr_to_mps(wind_speed_20ft)                     # m/s
            jd                 : JumpDistribution = resolve_JumpDistribution(spot_config,
                                                                             fireline_intensity,
                                                                             wind_speed_20ft_mps)

            #=======================================================================================
            # Cast each firebrand and accumulate any ignited cells
            #=======================================================================================

            # FIXME: A set is slow, so use a list instead. Collisions can happen with other source cells anyway.
            ignited_cells: set = set()
            i: cy.longlong
            for i in range(num_firebrands):
                ignited_cell: coord_yx = cast_firebrand(random_generator,
                                                        fuel_model_cube,
                                                        temperature_cube,
                                                        fuel_moisture_dead_1hr_cube,
                                                        fire_type_matrix,
                                                        rows,
                                                        cols,
                                                        cell_height,
                                                        cell_width,
                                                        t,
                                                        y,
                                                        x,
                                                        decay_distance,
                                                        cos_wdir,
                                                        sin_wdir,
                                                        jd)
                if (ignited_cell[0] != y) or (ignited_cell[1] != x):
                    ignited_cells.add(ignited_cell)

            #=======================================================================================
            # Return any cells ignited by firebrands along with their time of ignition
            #=======================================================================================

            if len(ignited_cells) > 0:
                ignition_time: cy.float = spot_ignition_time(time_of_arrival, flame_length) # minutes
                return (ignition_time, ignited_cells)
# spread-firebrands ends here



# [[file:../../org/pyretechnics.org::crown-fire-imports][crown-fire-imports]]
import cython
import cython as cy
if cython.compiled:
    from cython.cimports.libc.math import sqrt, exp, pow
    from cython.cimports.pyretechnics.cy_types import \
        vec_xyz, ProjectedVectors, FireBehaviorMax, SpreadBehavior, CrownSpreadInfo
    import cython.cimports.pyretechnics.conversion as conv
    import cython.cimports.pyretechnics.vector_utils as vu
    import cython.cimports.pyretechnics.surface_fire as sf
else:
    from math import sqrt, exp, pow
    from pyretechnics.py_types import \
        vec_xyz, ProjectedVectors, FireBehaviorMax, SpreadBehavior, CrownSpreadInfo
    import pyretechnics.conversion as conv
    import pyretechnics.vector_utils as vu
    import pyretechnics.surface_fire as sf
# crown-fire-imports ends here
# [[file:../../org/pyretechnics.org::van-wagner-critical-fireline-intensity][van-wagner-critical-fireline-intensity]]
@cy.cfunc
@cy.exceptval(check=False)
def van_wagner_critical_fireline_intensity(canopy_base_height: cy.float, foliar_moisture: cy.float) -> cy.float:
    """
    Returns the critical fireline intensity (kW/m) given:
    - canopy_base_height :: m
    - foliar_moisture    :: kg moisture/kg ovendry weight

    Constants used:
    460.0 = heat-of-ignition :: kJ/kg
    0.01 = empirical estimate for C in Van Wagner 1977 (eq. 4)
    """
    H: cy.float = 460.0 + 2600.0 * foliar_moisture
    v: cy.float = 0.01 * canopy_base_height * H
    return v * sqrt(v) # NOTE: This is faster than pow(v, 1.5).
# van-wagner-critical-fireline-intensity ends here
# [[file:../../org/pyretechnics.org::van-wagner-crowning-spread-rate][van-wagner-crowning-spread-rate]]
@cy.cfunc
@cy.exceptval(check=False)
def van_wagner_crowning_spread_rate(surface_fire_max  : FireBehaviorMax,
                                    canopy_base_height: cy.float,
                                    foliar_moisture   : cy.float) -> cy.float:
    """
    Returns the surface spread rate above which crown fire occurs (m/min) given:
    - surface_fire_max   :: FireBehaviorMax struct
    - canopy_base_height :: m
    - foliar_moisture    :: kg moisture/kg ovendry weight
    """
    surface_max_fireline_intensity: cy.float = surface_fire_max.max_fireline_intensity
    if surface_max_fireline_intensity > 0.0:
        surface_max_spread_rate    : cy.float = surface_fire_max.max_spread_rate
        critical_fireline_intensity: cy.float = van_wagner_critical_fireline_intensity(canopy_base_height,
                                                                                       foliar_moisture)
        return (surface_max_spread_rate * critical_fireline_intensity / surface_max_fireline_intensity)
    else:
        return 0.0
# van-wagner-crowning-spread-rate ends here
# [[file:../../org/pyretechnics.org::van-wagner-crown-fire-initiation][van-wagner-crown-fire-initiation]]
@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def van_wagner_crown_fire_initiation(surface_fireline_intensity: cy.float,
                                     canopy_cover              : cy.float,
                                     canopy_base_height        : cy.float,
                                     foliar_moisture           : cy.float) -> cy.bint:
    """
    Returns True if the surface fire transitions to a crown fire or False otherwise given:
    - surface_fireline_intensity :: kW/m
    - canopy_cover               :: 0-1
    - canopy_base_height         :: m
    - foliar_moisture            :: kg moisture/kg ovendry weight
    """
    return (
        surface_fireline_intensity > 0.0
        and
        canopy_cover > 0.4
        and
        surface_fireline_intensity >= van_wagner_critical_fireline_intensity(canopy_base_height, foliar_moisture)
    )
# van-wagner-crown-fire-initiation ends here
# [[file:../../org/pyretechnics.org::cruz-active-crown-fire-spread-rate][cruz-active-crown-fire-spread-rate]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def cruz_active_crown_fire_spread_rate(wind_speed_10m              : cy.float,
                                       canopy_bulk_density         : cy.float,
                                       estimated_fine_fuel_moisture: cy.float) -> cy.float:
    """
    Returns the active crown fire spread rate (m/min) given:
    - wind_speed_10m                                   :: km/hr
    - canopy_bulk_density                              :: kg/m^3
    - estimated_fine_fuel_moisture (M_f[0] "dead-1hr") :: kg moisture/kg ovendry weight
    """
    return (11.02
            * pow(wind_speed_10m, 0.90)
            * pow(canopy_bulk_density, 0.19)
            * exp(-17.0 * estimated_fine_fuel_moisture))
# cruz-active-crown-fire-spread-rate ends here
# [[file:../../org/pyretechnics.org::van-wagner-critical-spread-rate][van-wagner-critical-spread-rate]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def van_wagner_critical_spread_rate(canopy_bulk_density: cy.float) -> cy.float:
    """
    Returns the critical spread rate (m/min) given:
    - canopy_bulk_density :: kg/m^3
    """
    return 3.0 / canopy_bulk_density
# van-wagner-critical-spread-rate ends here
# [[file:../../org/pyretechnics.org::cruz-passive-crown-fire-spread-rate][cruz-passive-crown-fire-spread-rate]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def cruz_passive_crown_fire_spread_rate(active_spread_rate: cy.float, critical_spread_rate: cy.float) -> cy.float:
    """
    Returns the passive crown fire spread rate (m/min) given:
    - active_spread_rate   :: m/min
    - critical_spread_rate :: m/min
    """
    return active_spread_rate * exp(-active_spread_rate / critical_spread_rate)
# cruz-passive-crown-fire-spread-rate ends here
# [[file:../../org/pyretechnics.org::cruz-crown-fire-spread-info][cruz-crown-fire-spread-info]]
@cy.cfunc
@cy.exceptval(check=False)
def cruz_crown_fire_spread_info(wind_speed_10m              : cy.float,
                                canopy_bulk_density         : cy.float,
                                estimated_fine_fuel_moisture: cy.float) -> CrownSpreadInfo:
    """
    Given these inputs:
    - wind_speed_10m                                   :: km/hr
    - canopy_bulk_density                              :: kg/m^3
    - estimated_fine_fuel_moisture (M_f[0] "dead-1hr") :: kg moisture/kg ovendry weight

    return a CrownSpreadInfo struct containing these keys:
    - fire_type            :: 2 (passive_crown) or 3 (active_crown)
    - spread_rate          :: m/min
    - critical_spread_rate :: m/min
    """
    active_spread_rate  : cy.float = cruz_active_crown_fire_spread_rate(wind_speed_10m,
                                                                        canopy_bulk_density,
                                                                        estimated_fine_fuel_moisture) # m/min
    critical_spread_rate: cy.float = van_wagner_critical_spread_rate(canopy_bulk_density) # m/min
    if (active_spread_rate > critical_spread_rate):
        return CrownSpreadInfo(
            fire_type            = 3, # FIXME: NAMED CONSTANT
            spread_rate          = active_spread_rate,
            critical_spread_rate = critical_spread_rate,
        )
    else:
        return CrownSpreadInfo(
            fire_type            = 2, # FIXME: NAMED CONSTANT
            spread_rate          = cruz_passive_crown_fire_spread_rate(active_spread_rate, critical_spread_rate),
            critical_spread_rate = critical_spread_rate,
        )
# cruz-crown-fire-spread-info ends here
# [[file:../../org/pyretechnics.org::crown-fireline-intensity][crown-fireline-intensity]]
# NOTE: heat_of_combustion is h from the fuel models (generally 8000 Btu/lb).
# NOTE: ELMFIRE hard-codes heat_of_combustion to 18000 kJ/kg = 7738.6 Btu/lb.
@cy.cfunc
@cy.exceptval(check=False)
def calc_crown_fireline_intensity(crown_spread_rate  : cy.float,
                                  canopy_bulk_density: cy.float,
                                  canopy_height      : cy.float,
                                  canopy_base_height : cy.float,
                                  heat_of_combustion : cy.float) -> cy.float:
    """
    Returns the crown fireline intensity (Btu/ft/s OR kW/m) given:
    - crown_spread_rate   :: ft/min  OR m/min
    - canopy_bulk_density :: lb/ft^3 OR kg/m^3
    - canopy_height       :: ft      OR m
    - canopy_base_height  :: ft      OR m
    - heat_of_combustion  :: Btu/lb  OR kJ/kg

    (ft/min * lb/ft^3 * ft * Btu/lb)/60 = (Btu/ft/min)/60 = Btu/ft/s
    OR
    (m/min * kg/m^3 * m * kJ/kg)/60 = (kJ/m*min)/60 = kJ/m*s = kW/m
    """
    canopy_height_difference: cy.float = canopy_height - canopy_base_height
    return (crown_spread_rate * canopy_bulk_density * canopy_height_difference * heat_of_combustion) / 60.0
# crown-fireline-intensity ends here
# [[file:../../org/pyretechnics.org::crown-fire-eccentricity][crown-fire-eccentricity]]
# Parameters for the linear model that computes LoW from wind speed.
@cy.cfunc
@cy.exceptval(check=False)
def crown_length_to_width_ratio(wind_speed_10m: cy.float, max_length_to_width_ratio: cy.float = 1e10) -> cy.float:
    """
    Calculate the length_to_width_ratio of the crown fire front using eq. 9 from
    Rothermel 1991 given:
    - wind_speed_10m            :: km/hr (aligned with the slope-tangential plane)
    - max_length_to_width_ratio :: float > 0.0 (Optional)
    """
    wind_speed_20ft      : cy.float = conv.wind_speed_10m_to_wind_speed_20ft(wind_speed_10m) # km/hr
    wind_speed_20ft_mph  : cy.float = conv.km_hr_to_mph(wind_speed_20ft)                     # mph
    length_to_width_ratio: cy.float = 1.0 + 0.125 * wind_speed_20ft_mph
    return min(length_to_width_ratio, max_length_to_width_ratio)


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def crown_fire_eccentricity(length_to_width_ratio: cy.float) -> cy.float:
    """
    Calculate the eccentricity (E) of the crown fire front using eq. 8 from
    Albini and Chase 1980 given:
    - length_to_width_ratio :: (1: circular spread, > 1: elliptical spread)
    """
    return sqrt(length_to_width_ratio * length_to_width_ratio - 1.0) / length_to_width_ratio
# crown-fire-eccentricity ends here
# [[file:../../org/pyretechnics.org::crown-fire-behavior-max][crown-fire-behavior-max]]
@cy.ccall
@cy.exceptval(check=False)
def calc_crown_fire_behavior_max(canopy_height               : cy.float,
                                 canopy_base_height          : cy.float,
                                 canopy_bulk_density         : cy.float,
                                 heat_of_combustion          : cy.float,
                                 estimated_fine_fuel_moisture: cy.float,
                                 wind_speed_10m              : cy.float,
                                 upwind_direction            : cy.float,
                                 slope                       : cy.float,
                                 aspect                      : cy.float,
                                 crown_max_lw_ratio          : cy.float=1e10) -> FireBehaviorMax:
    """
    Given these inputs:
    - canopy_height                                    :: m
    - canopy_base_height                               :: m
    - canopy_bulk_density                              :: kg/m^3
    - heat_of_combustion                               :: kJ/kg
    - estimated_fine_fuel_moisture (M_f[0] "dead-1hr") :: kg moisture/kg ovendry weight
    - wind_speed_10m                                   :: km/hr
    - upwind_direction                                 :: degrees clockwise from North
    - slope                                            :: rise/run
    - aspect                                           :: degrees clockwise from North
    - crown_max_lw_ratio                               :: float > 0.0 (Optional)

    return a FireBehaviorMax struct containing these keys:
    - max_fire_type          :: 2 (passive_crown) or 3 (active_crown)
    - max_spread_rate        :: m/min
    - max_spread_direction   :: (x, y, z) unit vector
    - max_fireline_intensity :: kW/m
    - max_flame_length       :: m
    - length_to_width_ratio  :: unitless (1: circular spread, > 1: elliptical spread)
    - eccentricity           :: unitless (0: circular spread, > 0: elliptical spread)
    - critical_spread_rate   :: m/min
    """
    # Reverse the provided wind and slope directions
    downwind_direction: cy.float = conv.opposite_direction(upwind_direction)
    upslope_direction : cy.float = conv.opposite_direction(aspect)
    # Project wind and slope vectors onto the slope-tangential plane
    # FIXME: Let's just have these vectors as arguments to the function instead of re-computing them.
    vectors: ProjectedVectors = sf.project_wind_and_slope_vectors_3d(wind_speed_10m, downwind_direction,
                                                                     slope, upslope_direction)
    wind_vector_3d : vec_xyz = vectors.wind_vector_3d  # km/hr
    slope_vector_3d: vec_xyz = vectors.slope_vector_3d # rise/run
    # Determine the max spread direction
    # FIXME: REVIEW Should we make the max_spread_direction the combined wind and slope direction?
    wind_speed_10m_3d: cy.float = vu.vector_magnitude_3d(wind_vector_3d) # km/hr
    max_spread_direction: vec_xyz
    if wind_speed_10m_3d > 0.0:
        max_spread_direction = vu.as_unit_vector_3d(wind_vector_3d)  # unit vector in the 3D downwind direction
    elif slope > 0.0:
        max_spread_direction = vu.as_unit_vector_3d(slope_vector_3d) # unit vector in the 3D upslope direction
    else:
        max_spread_direction = (0.0, 1.0, 0.0)                       # default: North
    # Calculate the crown fire behavior in the max spread direction
    spread_info          : CrownSpreadInfo = cruz_crown_fire_spread_info(wind_speed_10m_3d,
                                                                         canopy_bulk_density,
                                                                         estimated_fine_fuel_moisture)
    spread_rate          : cy.float        = spread_info.spread_rate                           # m/min
    fireline_intensity   : cy.float        = calc_crown_fireline_intensity(spread_rate,
                                                                           canopy_bulk_density,
                                                                           canopy_height,
                                                                           canopy_base_height,
                                                                           heat_of_combustion) # kW/m
    length_to_width_ratio: cy.float        = crown_length_to_width_ratio(wind_speed_10m_3d,
                                                                         crown_max_lw_ratio)   # unitless
    eccentricity         : cy.float        = crown_fire_eccentricity(length_to_width_ratio)    # unitless
    return FireBehaviorMax(
        max_fire_type          = spread_info.fire_type,
        max_spread_rate        = spread_rate,
        max_spread_direction   = max_spread_direction, # unit vector
        max_fireline_intensity = fireline_intensity,
        max_flame_length       = 0.0, # NOTE: max_flame_length is not provided, as in the original unoptimized code.
        length_to_width_ratio  = length_to_width_ratio,
        eccentricity           = eccentricity,
        critical_spread_rate   = spread_info.critical_spread_rate,
    )
# crown-fire-behavior-max ends here
# [[file:../../org/pyretechnics.org::crown-fire-behavior-in-direction][crown-fire-behavior-in-direction]]
@cy.ccall
@cy.exceptval(check=False)
def calc_crown_fire_behavior_in_direction(crown_fire_max  : FireBehaviorMax,
                                          spread_direction: vec_xyz) -> SpreadBehavior:
    """
    Given these inputs:
    - crown_fire_max     :: a FireBehaviorMax struct of max crown fire behavior values
      - max_fire_type          :: 2 (passive_crown) or 3 (active_crown)
      - max_spread_rate        :: m/min
      - max_spread_direction   :: (x, y, z) unit vector
      - max_fireline_intensity :: kW/m
      - max_flame_length       :: m
      - length_to_width_ratio  :: unitless (1: circular spread, > 1: elliptical spread)
      - eccentricity           :: unitless (0: circular spread, > 0: elliptical spread)
      - critical_spread_rate   :: m/min
    - spread_direction   :: 3D unit vector on the slope-tangential plane

    return a SpreadBehavior struct containing these keys:
    - dphi_dt            :: phi/min
    - fire_type          :: 2 (passive_crown) or 3 (active_crown)
    - spread_rate        :: m/min
    - spread_direction   :: (x, y, z) unit vector
    - fireline_intensity :: kW/m
    - flame_length       :: m
    """
    # Unpack max crown fire behavior values
    max_fire_type         : cy.int   = crown_fire_max.max_fire_type
    max_spread_rate       : cy.float = crown_fire_max.max_spread_rate
    max_spread_direction  : vec_xyz  = crown_fire_max.max_spread_direction
    max_fireline_intensity: cy.float = crown_fire_max.max_fireline_intensity
    eccentricity          : cy.float = crown_fire_max.eccentricity
    critical_spread_rate  : cy.float = crown_fire_max.critical_spread_rate
    # Calculate cos(w), where w is the offset angle between these unit vectors on the slope-tangential plane
    cos_w: cy.float = vu.dot_3d(max_spread_direction, spread_direction)
    # Calculate adjustment due to the offset angle from the max spread direction
    adjustment: cy.float = (1.0 - eccentricity) / (1.0 - eccentricity * cos_w)
    # Adjust the spread rate (possibly switching from an active to passive crown fire)
    spread_rate: cy.float = max_spread_rate * adjustment
    if spread_rate > critical_spread_rate:
        # Max spread rate was active and directional spread rate remains active
        return SpreadBehavior(
            dphi_dt            = 0.0,
            fire_type          = 3, # active_crown
            spread_rate        = spread_rate,
            spread_direction   = spread_direction,
            fireline_intensity = max_fireline_intensity * adjustment,
            flame_length       = 0.0,
        )
    elif max_fire_type == 2: # passive_crown
        # Max spread rate was passive and directional spread rate remains passive
        return SpreadBehavior(
            dphi_dt            = 0.0,
            fire_type          = 2, # passive_crown
            spread_rate        = spread_rate,
            spread_direction   = spread_direction,
            fireline_intensity = max_fireline_intensity * adjustment,
            flame_length       = 0.0,
        )
    else:
        # Max spread rate was active and directional spread rate has become passive
        return SpreadBehavior(
            dphi_dt            = 0.0,
            fire_type          = 2, # passive_crown
            spread_rate        = cruz_passive_crown_fire_spread_rate(spread_rate, critical_spread_rate),
            spread_direction   = spread_direction,
            fireline_intensity = max_fireline_intensity * adjustment,
            flame_length       = 0.0,
        )
# crown-fire-behavior-in-direction ends here
# [[file:../../org/pyretechnics.org::combined-fire-behavior][combined-fire-behavior]]
@cy.ccall
@cy.exceptval(check=False)
def calc_combined_fire_behavior(surface_fire_behavior: SpreadBehavior,
                                crown_fire_behavior  : SpreadBehavior) -> SpreadBehavior:
    """
    Given these inputs:
    - surface_fire_behavior :: a SpreadBehavior struct of surface fire behavior values
      - dphi_dt                :: phi/min
      - fire_type              :: 1 (surface)
      - spread_rate            :: m/min
      - spread_direction       :: (x, y, z) unit vector
      - fireline_intensity     :: kW/m
      - flame_length           :: m
    - crown_fire_behavior   :: a SpreadBehavior struct of crown fire behavior values
      - dphi_dt                :: phi/min
      - fire_type              :: 2 (passive_crown) or 3 (active_crown)
      - spread_rate            :: m/min
      - spread_direction       :: (x, y, z) unit vector
      - fireline_intensity     :: kW/m
      - flame_length           :: m

    return a SpreadBehavior struct containing these keys:
    - dphi_dt            :: phi/min
    - fire_type          :: 1 (surface), 2 (passive_crown), or 3 (active_crown)
    - spread_rate        :: m/min
    - spread_direction   :: (x, y, z) unit vector
    - fireline_intensity :: kW/m
    - flame_length       :: m
    """
    # Unpack the surface fire behavior values
    surface_dphi_dt           : cy.float = surface_fire_behavior.dphi_dt            # phi/min
    surface_spread_rate       : cy.float = surface_fire_behavior.spread_rate        # m/min
    surface_spread_direction  : vec_xyz  = surface_fire_behavior.spread_direction   # (x, y, z) unit vector
    surface_fireline_intensity: cy.float = surface_fire_behavior.fireline_intensity # kW/m
    surface_flame_length      : cy.float = surface_fire_behavior.flame_length       # m
    # Unpack the crown fire behavior values
    crown_dphi_dt           : cy.float = crown_fire_behavior.dphi_dt            # phi/min
    crown_fire_type         : cy.int   = crown_fire_behavior.fire_type          # 2 (passive_crown) or 3 (active_crown)
    crown_spread_rate       : cy.float = crown_fire_behavior.spread_rate        # m/min
    crown_spread_direction  : vec_xyz  = crown_fire_behavior.spread_direction   # (x, y, z) unit vector
    crown_fireline_intensity: cy.float = crown_fire_behavior.fireline_intensity # kW/m
    # Select the most quickly changing (i.e., lowest) dphi_dt value
    dphi_dt: cy.float = min(surface_dphi_dt, crown_dphi_dt)
    # Determine whether the surface or crown fire has the fastest spread rate
    if surface_spread_rate == 0.0:
        # Independent crown fire (NOTE: This is probably user error.)
        return crown_fire_behavior
    elif crown_spread_rate == 0.0:
        if crown_fire_type == 2:
            # Passive crown fire
            return SpreadBehavior(
                dphi_dt            = dphi_dt,
                fire_type          = crown_fire_type,
                spread_rate        = surface_spread_rate,
                spread_direction   = surface_spread_direction,
                fireline_intensity = surface_fireline_intensity,
                flame_length       = surface_flame_length,
            )
        else:
            # No crown fire
            return surface_fire_behavior
    elif surface_spread_rate > crown_spread_rate:
        # Surface fire spreads faster
        combined_fireline_intensity: cy.float = (surface_fireline_intensity
                                                 + crown_fireline_intensity * surface_spread_rate / crown_spread_rate)
        return SpreadBehavior(
            dphi_dt            = dphi_dt,
            fire_type          = crown_fire_type,
            spread_rate        = surface_spread_rate,
            spread_direction   = surface_spread_direction,
            fireline_intensity = combined_fireline_intensity,
            flame_length       = sf.calc_flame_length(combined_fireline_intensity),
        )
    else:
        # Crown fire spreads faster
        combined_fireline_intensity: cy.float = (surface_fireline_intensity * crown_spread_rate / surface_spread_rate
                                                 + crown_fireline_intensity)
        return SpreadBehavior(
            dphi_dt            = dphi_dt,
            fire_type          = crown_fire_type,
            spread_rate        = crown_spread_rate,
            spread_direction   = crown_spread_direction,
            fireline_intensity = combined_fireline_intensity,
            flame_length       = sf.calc_flame_length(combined_fireline_intensity),
        )
# combined-fire-behavior ends here



# [[file:../../org/pyretechnics.org::space-time-cube-imports][space-time-cube-imports]]
import cython
import cython as cy
from functools import reduce
import numpy as np
if cython.compiled:
    from cython.cimports.numpy import ndarray
    from cython.cimports.pyretechnics.cy_types import pyidx
else:
    from numpy import ndarray
    from pyretechnics.py_types import pyidx
# space-time-cube-imports ends here
# [[file:../../org/pyretechnics.org::space-time-cube-utilities][space-time-cube-utilities]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def is_pos_int(x: object) -> cy.bint:
    return isinstance(x, int) and x > 0


@cy.cfunc
def divide_evenly(dividend: cy.int, divisor: cy.int) -> cy.int:
    if divisor == 0:
        raise ValueError(f"{divisor} may not be zero.")
    else:
        quotient : cy.int = dividend // divisor
        remainder: cy.int = dividend % divisor
        if remainder == 0:
            return quotient
        else:
            raise ValueError(f"{dividend} must be an exact multiple of {divisor}.")


@cy.ccall
@cy.exceptval(check=False)
def to_positive_index_range(index_range: tuple[pyidx, pyidx]|None, axis_length: pyidx) -> tuple[pyidx, pyidx]:
    """
    Translate None and negative indices to positive indices.
    """
    if index_range is None:
        return (0, axis_length)
    else:
        start: pyidx|None = index_range[0]
        stop : pyidx|None = index_range[1]
        return (
            0 if start is None else axis_length + start if start < 0 else start,
            axis_length if stop is None else axis_length + stop if stop < 0 else stop
        )


@cy.cfunc
def stretch_array(old_array: ndarray, new_length: cy.int, repetitions: cy.float) -> ndarray:
    new_array: ndarray = np.zeros(new_length, dtype=old_array.dtype)
    i        : pyidx
    for i in range(new_length):
        new_array[i] = old_array[int(i / repetitions)]
    return new_array


@cy.cfunc
def maybe_repeat_array(maybe_array: ndarray, axis_repetitions: tuple[pyidx, cy.float]) -> ndarray:
    """
    Return a new array that is created by repeating the elements from the input
    array repetitions times along the specified array axis. Avoid allocating
    new memory if repetitions == 1 or if the repeated array axis has length 1.
    """
    axis       : pyidx    = axis_repetitions[0]
    repetitions: cy.float = axis_repetitions[1]
    old_array  : ndarray  = np.asarray(maybe_array)
    array_dims : pyidx    = old_array.ndim
    array_shape: list     = list(np.shape(old_array))
    axis_length: cy.int   = array_shape[axis]
    if repetitions == 1.0:
        # no repetitions necessary
        return old_array
    elif axis_length == 1:
        # broadcast single-element axis repetitions times
        array_shape[axis] = int(repetitions)
        return np.broadcast_to(old_array, array_shape)
    elif repetitions % 1.0 == 0.0:
        # repeat each element on the chosen axis repetitions times
        return np.repeat(old_array, int(repetitions), axis)
    elif axis == 0 and array_dims == 1:
        # populate a new 1D array of the expected length by translating its indices into the original array
        new_axis_length: cy.int = int(axis_length * repetitions)
        return stretch_array(old_array, new_axis_length, repetitions)
    elif axis == 0 and array_dims == 2:
        # populate a new 2D array of the expected length by translating its indices into the original array
        new_rows: cy.int = int(axis_length * repetitions)
        return np.stack([old_array[int(i / repetitions)] for i in range(new_rows)])
    elif axis == 1 and array_dims == 2:
        # populate a new 2D array of the expected length by translating its indices into the original array
        old_rows: cy.int = array_shape[0]
        new_cols: cy.int = int(axis_length * repetitions)
        return np.stack([stretch_array(old_array[i], new_cols, repetitions) for i in range(old_rows)])
    elif axis == 0 and array_dims == 3:
        # populate a new 3D array of the expected length by translating its indices into the original array
        new_bands: cy.int = int(axis_length * repetitions)
        return np.stack([old_array[int(b / repetitions)] for b in range(new_bands)])
    elif axis == 1 and array_dims == 3:
        # populate a new 2D array of the expected length by translating its indices into the original array
        old_bands: cy.int = array_shape[0]
        new_rows : cy.int = int(axis_length * repetitions)
        return np.stack([
            np.stack([
                old_array[b, int(i / repetitions)]
                for i in range(new_rows)
            ])
            for b in range(old_bands)
        ])
    elif axis == 2 and array_dims == 3:
        # populate a new 3D array of the expected length by translating its indices into the original array
        old_bands: cy.int = array_shape[0]
        old_rows : cy.int = array_shape[1]
        new_cols : cy.int = int(axis_length * repetitions)
        return np.stack([
            np.stack([
                stretch_array(old_array[b,i], new_cols, repetitions)
                for i in range(old_rows)
            ])
            for b in range(old_bands)
        ])
    else:
        raise ValueError("Floating point repetitions are only supported for 1D, 2D, and 3D arrays.")
# space-time-cube-utilities ends here
# [[file:../../org/pyretechnics.org::ispace-time-cube-class][ispace-time-cube-class]]
@cy.cclass
class ISpaceTimeCube:
    @cy.cfunc
    @cy.exceptval(check=False)
    def get(self, t: pyidx, y: pyidx, x: pyidx) -> cy.float:
        pass
# ispace-time-cube-class ends here
# [[file:../../org/pyretechnics.org::space-time-cube-class][space-time-cube-class]]
@cy.cclass
class SpaceTimeCube(ISpaceTimeCube):
    """
    Create an object that represents a 3D array with dimensions (T,Y,X) given by cube_shape.
    Internally, data is stored as a 3D Numpy array at the resolution of the provided base data.
    Whenever a point value or contiguous space-time region of values is requested, translate
    the given cube_shape coordinates into base coordinates, look up the values from the base data,
    expand them (if necessary) back into the cube_shape resolution, and return the resulting scalar
    value or array to the caller.
    """
    ndim         : cy.int
    size         : cy.ulonglong
    shape        : tuple[cy.int, cy.int, cy.int]
    base         : object
    t_repetitions: cy.float
    y_repetitions: cy.float
    x_repetitions: cy.float
    data         : cy.float[:,:,::1] # FIXME: Restore polymorphism for the underlying Numpy arrays


    def __init__(self, cube_shape: tuple[int, int, int], base: object) -> cy.void:
        """
        NOTE: The resolutions in cube_shape must be exact multiples of any existing dimensions
              in the base data. If base is not a Numpy float32 array, a new array will be allocated.
        """
        # Ensure that cube_shape contains 3 values or throw an error
        if len(cube_shape) != 3:
            raise ValueError("The cube_shape must contain exactly three values.")

        # Unpack the cube_shape values without type-checking
        cube_bands_: object = cube_shape[0]
        cube_rows_ : object = cube_shape[1]
        cube_cols_ : object = cube_shape[2]

        # Ensure that cube_shape only contains positive integers or throw an error
        if not(is_pos_int(cube_bands_) and is_pos_int(cube_rows_) and is_pos_int(cube_cols_)):
            raise ValueError("The cube_shape must contain only positive integers.")

        # Cast the cube_shape values as primitive ints
        cube_bands: cy.int = cube_bands_
        cube_rows : cy.int = cube_rows_
        cube_cols : cy.int = cube_cols_

        # Store the cube metadata for later
        self.ndim  = 3
        self.size  = cube_bands * cube_rows * cube_cols
        self.shape = (cube_bands, cube_rows, cube_cols)
        self.base  = base

        # Store the base data as a 3D array along with its axis repetitions
        base_dimensions: cy.int = np.ndim(base)

        if base_dimensions == 0:
            # 0D: Constant Input
            self.t_repetitions = cube_bands
            self.y_repetitions = cube_rows
            self.x_repetitions = cube_cols
            self.data          = np.asarray([[[base]]], dtype=np.float32)

        elif base_dimensions == 1:
            # 1D: Time-Series Input
            base_bands: cy.int = len(base)
            self.t_repetitions = float(cube_bands) / float(base_bands)
            self.y_repetitions = cube_rows
            self.x_repetitions = cube_cols
            # Ensure that the cube_shape is not smaller than the base shape
            if cube_bands < base_bands:
                raise ValueError("The cube_shape may not be smaller than the base shape.")
            # Warn if any repetitions are not whole numbers
            if self.t_repetitions % 1.0 != 0.0:
                print("WARNING: Input data's shape does not evenly divide the cube_shape."
                      + " Index lookups beyond the cube's edge may mistakenly return values without errors.")
            # Warn if base is not a Numpy float32 array
            if not(isinstance(base, np.ndarray)) or (base.dtype != np.float32):
                print("WARNING: Input data is not a Numpy float32 array. Data will be copied into SpaceTimeCube.",
                      flush=True)
            # Expand (base_bands) -> (base_bands,1,1)
            self.data = np.expand_dims(np.asarray(base, dtype=np.float32), axis=(1,2))

        elif base_dimensions == 2:
            # 2D: Spatial Input
            base_shape: tuple  = np.shape(base)
            base_rows : cy.int = base_shape[0]
            base_cols : cy.int = base_shape[1]
            self.t_repetitions = cube_bands
            self.y_repetitions = float(cube_rows) / float(base_rows)
            self.x_repetitions = float(cube_cols) / float(base_cols)
            # Ensure that the cube_shape is not smaller than the base shape
            if cube_rows < base_rows or cube_cols < base_cols:
                raise ValueError("The cube_shape may not be smaller than the base shape.")
            # Warn if any repetitions are not whole numbers
            if self.y_repetitions % 1.0 != 0.0 or self.x_repetitions % 1.0 != 0.0:
                print("WARNING: Input data's shape does not evenly divide the cube_shape."
                      + " Index lookups beyond the cube's edge may mistakenly return values without errors.")
            # Warn if base is not a Numpy float32 array
            if not(isinstance(base, np.ndarray)) or (base.dtype != np.float32):
                print("WARNING: Input data is not a Numpy float32 array. Data will be copied into SpaceTimeCube.",
                      flush=True)
            # Expand (base_rows,base_cols) -> (1,base_rows,base_cols)
            self.data = np.expand_dims(np.asarray(base, dtype=np.float32), axis=0)

        elif base_dimensions == 3:
            # 3D: Spatio-Temporal Input
            base_shape: tuple  = np.shape(base)
            base_bands: cy.int = base_shape[0]
            base_rows : cy.int = base_shape[1]
            base_cols : cy.int = base_shape[2]
            self.t_repetitions = float(cube_bands) / float(base_bands)
            self.y_repetitions = float(cube_rows) / float(base_rows)
            self.x_repetitions = float(cube_cols) / float(base_cols)
            # Ensure that the cube_shape is not smaller than the base shape
            if cube_bands < base_bands or cube_rows < base_rows or cube_cols < base_cols:
                raise ValueError("The cube_shape may not be smaller than the base shape.")
            # Warn if any repetitions are not whole numbers
            if self.t_repetitions % 1.0 != 0.0 or self.y_repetitions % 1.0 != 0.0 or self.x_repetitions % 1.0 != 0.0:
                print("WARNING: Input data's shape does not evenly divide the cube_shape."
                      + " Index lookups beyond the cube's edge may mistakenly return values without errors.")
            # Warn if base is not a Numpy float32 array
            if not(isinstance(base, np.ndarray)) or (base.dtype != np.float32):
                print("WARNING: Input data is not a Numpy float32 array. Data will be copied into SpaceTimeCube.",
                      flush=True)
            self.data = np.asarray(base, dtype=np.float32)

        else:
            # 4D+: Invalid Input
            raise ValueError("Invalid input: base must have 0-3 dimensions.")


    @cy.ccall
    @cy.exceptval(check=False)
    def get(self, t: pyidx, y: pyidx, x: pyidx) -> cy.float:
        """
        Return the scalar value at index (t,y,x) by translating these cube coordinates
        to base coordinates and looking up the value within the base data.

        NOTE: Indices may be negative.
        """
        # Select value by spatio-temporal coordinate
        base_t: pyidx = int(t / self.t_repetitions)
        base_y: pyidx = int(y / self.y_repetitions)
        base_x: pyidx = int(x / self.x_repetitions)
        return self.data[base_t, base_y, base_x]


    @cy.ccall
    def getTimeSeries(self, t_range: tuple[pyidx, pyidx]|None, y: pyidx, x: pyidx) -> ndarray:
        """
        Return the 1D array given by the slice (t_range,y,x) by translating these cube
        coordinates to base coordinates, looking up the array slice within the base data,
        and expanding it back to the cube_shape resolution.

        NOTE: Indices may be negative.
        NOTE: Range indices may include one or more None values and
              provide (inclusion, exclusion) semantics like Python array slice notation.
        """
        # Destructure the argument range
        t_range_updated : tuple[pyidx, pyidx] = to_positive_index_range(t_range, self.shape[0])
        t_start         : pyidx               = t_range_updated[0]
        t_stop_exclusive: pyidx               = t_range_updated[1]
        t_stop          : pyidx               = t_stop_exclusive - 1
        # Translate high-res coordinates to low-res coordinates
        t_start_chunk: pyidx = int(t_start / self.t_repetitions)
        t_stop_chunk : pyidx = int(t_stop  / self.t_repetitions)
        y_chunk      : pyidx = int(y       / self.y_repetitions)
        x_chunk      : pyidx = int(x       / self.x_repetitions)
        # Select the array slice that completely contains all low-res coordinates
        low_res_time: ndarray = np.asarray(self.data[t_start_chunk:(t_stop_chunk + 1),
                                                     y_chunk,
                                                     x_chunk])
        # Expand the low-res slice into a high-res slice
        high_res_time: ndarray = maybe_repeat_array(low_res_time, (0, self.t_repetitions))
        # Translate high-res global coordinates to high-res slice coordinates
        t_chunk_origin: pyidx = int(t_start_chunk * self.t_repetitions)
        t_start_idx   : pyidx = t_start - t_chunk_origin
        t_stop_idx    : pyidx = t_stop  - t_chunk_origin
        # Select the array slice that matches the high-res slice coordinates
        return high_res_time[t_start_idx:(t_stop_idx + 1)]


    @cy.ccall
    def getSpatialPlane(self,
                        t      : pyidx,
                        y_range: tuple[pyidx, pyidx]|None,
                        x_range: tuple[pyidx, pyidx]|None) -> ndarray:
        """
        Return the 2D array given by the slice (t,y_range,x_range) by translating these
        cube coordinates to base coordinates, looking up the array slice within the base
        data, and expanding it back to the cube_shape resolution.

        NOTE: Indices may be negative.
        NOTE: Range indices may include one or more None values and
              provide (inclusion, exclusion) semantics like Python array slice notation.
        """
        # Destructure the argument ranges
        y_range_updated : tuple[pyidx, pyidx] = to_positive_index_range(y_range, self.shape[1])
        x_range_updated : tuple[pyidx, pyidx] = to_positive_index_range(x_range, self.shape[2])
        y_start         : pyidx               = y_range_updated[0]
        y_stop_exclusive: pyidx               = y_range_updated[1]
        x_start         : pyidx               = x_range_updated[0]
        x_stop_exclusive: pyidx               = x_range_updated[1]
        y_stop          : pyidx               = y_stop_exclusive - 1
        x_stop          : pyidx               = x_stop_exclusive - 1
        # Translate high-res coordinates to low-res coordinates
        t_chunk      : pyidx = int(t       / self.t_repetitions)
        y_start_chunk: pyidx = int(y_start / self.y_repetitions)
        y_stop_chunk : pyidx = int(y_stop  / self.y_repetitions)
        x_start_chunk: pyidx = int(x_start / self.x_repetitions)
        x_stop_chunk : pyidx = int(x_stop  / self.x_repetitions)
        # Select the array slice that completely contains all low-res coordinates
        low_res_space: ndarray = np.asarray(self.data[t_chunk,
                                                      y_start_chunk:(y_stop_chunk + 1),
                                                      x_start_chunk:(x_stop_chunk + 1)])
        # Expand the low-res slice into a high-res slice
        high_res_space: ndarray = reduce(maybe_repeat_array,
                                         ((0, self.y_repetitions),
                                          (1, self.x_repetitions)),
                                         low_res_space)
        # Translate high-res global coordinates to high-res slice coordinates
        y_chunk_origin: pyidx = int(y_start_chunk * self.y_repetitions)
        x_chunk_origin: pyidx = int(x_start_chunk * self.x_repetitions)
        y_start_idx   : pyidx = y_start - y_chunk_origin
        y_stop_idx    : pyidx = y_stop  - y_chunk_origin
        x_start_idx   : pyidx = x_start - x_chunk_origin
        x_stop_idx    : pyidx = x_stop  - x_chunk_origin
        # Select the array slice that matches the high-res slice coordinates
        return high_res_space[y_start_idx:(y_stop_idx + 1),
                              x_start_idx:(x_stop_idx + 1)]


    @cy.ccall
    def getSubcube(self,
                   t_range: tuple[pyidx, pyidx]|None,
                   y_range: tuple[pyidx, pyidx]|None,
                   x_range: tuple[pyidx, pyidx]|None) -> ndarray:
        """
        Return the 3D array given by the slice (t_range,y_range,x_range) by translating
        these cube coordinates to base coordinates, looking up the array slice within the
        base data, and expanding it back to the cube_shape resolution.

        NOTE: Indices may be negative.
        NOTE: Range indices may include one or more None values and
              provide (inclusion, exclusion) semantics like Python array slice notation.
        """
        # Destructure the argument ranges
        t_range_updated : tuple[pyidx, pyidx] = to_positive_index_range(t_range, self.shape[0])
        y_range_updated : tuple[pyidx, pyidx] = to_positive_index_range(y_range, self.shape[1])
        x_range_updated : tuple[pyidx, pyidx] = to_positive_index_range(x_range, self.shape[2])
        t_start         : pyidx               = t_range_updated[0]
        t_stop_exclusive: pyidx               = t_range_updated[1]
        y_start         : pyidx               = y_range_updated[0]
        y_stop_exclusive: pyidx               = y_range_updated[1]
        x_start         : pyidx               = x_range_updated[0]
        x_stop_exclusive: pyidx               = x_range_updated[1]
        t_stop          : pyidx               = t_stop_exclusive - 1
        y_stop          : pyidx               = y_stop_exclusive - 1
        x_stop          : pyidx               = x_stop_exclusive - 1
        # Translate high-res coordinates to low-res coordinates
        t_start_chunk: pyidx = int(t_start / self.t_repetitions)
        t_stop_chunk : pyidx = int(t_stop  / self.t_repetitions)
        y_start_chunk: pyidx = int(y_start / self.y_repetitions)
        y_stop_chunk : pyidx = int(y_stop  / self.y_repetitions)
        x_start_chunk: pyidx = int(x_start / self.x_repetitions)
        x_stop_chunk : pyidx = int(x_stop  / self.x_repetitions)
        # Select the array slice that completely contains all low-res coordinates
        low_res_cube: ndarray = np.asarray(self.data[t_start_chunk:(t_stop_chunk + 1),
                                                     y_start_chunk:(y_stop_chunk + 1),
                                                     x_start_chunk:(x_stop_chunk + 1)])
        # Expand the low-res slice into a high-res slice
        high_res_cube: ndarray = reduce(maybe_repeat_array,
                                        ((0, self.t_repetitions),
                                         (1, self.y_repetitions),
                                         (2, self.x_repetitions)),
                                        low_res_cube)
        # Translate high-res global coordinates to high-res slice coordinates
        t_chunk_origin: pyidx = int(t_start_chunk * self.t_repetitions)
        y_chunk_origin: pyidx = int(y_start_chunk * self.y_repetitions)
        x_chunk_origin: pyidx = int(x_start_chunk * self.x_repetitions)
        t_start_idx   : pyidx = t_start - t_chunk_origin
        t_stop_idx    : pyidx = t_stop  - t_chunk_origin
        y_start_idx   : pyidx = y_start - y_chunk_origin
        y_stop_idx    : pyidx = y_stop  - y_chunk_origin
        x_start_idx   : pyidx = x_start - x_chunk_origin
        x_stop_idx    : pyidx = x_stop  - x_chunk_origin
        # Select the array slice that matches the high-res slice coordinates
        return high_res_cube[t_start_idx:(t_stop_idx + 1),
                             y_start_idx:(y_stop_idx + 1),
                             x_start_idx:(x_stop_idx + 1)]


    def __getFullyRealizedCube(self):
        """
        Return the 3D array created by expanding the base data to the cube_shape resolution.
        Wherever possible, Numpy broadcasting is used to avoid memory allocation along
        constant array dimensions.
        """
        base_dimensions = np.ndim(self.base)

        if base_dimensions == 0:
            # 0D: Constant Input
            # Broadcast (0,0,0) -> (t,y,x)
            return np.broadcast_to(self.data, self.shape)

        elif base_dimensions == 1:
            # 1D: Time-Series Input
            # Repeat (t0,1,1) -> (t,1,1)
            repeated_array = maybe_repeat_array(np.asarray(self.data), (0, self.t_repetitions))
            # Broadcast (t,1,1) -> (t,y,x)
            return np.broadcast_to(repeated_array, self.shape)

        elif base_dimensions == 2:
            # 2D: Spatial Input
            # Repeat (1,y0,x0) -> (1,y,x)
            repeated_array = reduce(maybe_repeat_array,
                                    ((1, self.y_repetitions),
                                     (2, self.x_repetitions)),
                                    np.asarray(self.data))
            # Broadcast (1,y,x) -> (t,y,x)
            return np.broadcast_to(repeated_array, self.shape)

        else:
            # 3D: Spatio-Temporal Input
            # Repeat (t0,y0,x0) -> (t,y,x)
            return reduce(maybe_repeat_array,
                          ((0, self.t_repetitions),
                           (1, self.y_repetitions),
                           (2, self.x_repetitions)),
                          np.asarray(self.data))


    def getFullyRealizedCube(self, cache=False):
        """
        Return the 3D array created by expanding the base data to the cube_shape resolution.
        Wherever possible, Numpy broadcasting is used to avoid memory allocation along
        constant array dimensions. When cache == True, this expanded 3D array is cached
        within the SpaceTimeCube object for future immediate retrieval.
        """
        if hasattr(self, "cube"):
            return self.cube
        else:
            cube = self.__getFullyRealizedCube()
            if cache is True:
                self.cube = cube
            return cube


    def releaseFullyRealizedCube(self):
        """
        Deletes the cached fully realized cube if it exists.
        """
        if hasattr(self, "cube"):
            delattr(self, "cube")
# space-time-cube-class ends here
# [[file:../../org/pyretechnics.org::lazy-space-time-cube-class][lazy-space-time-cube-class]]
@cy.cclass
class LazySpaceTimeCube(ISpaceTimeCube):
    """
    Create an object that represents a 3D array with dimensions (T,Y,X) given by cube_shape.
    Internally, data is stored as an initially empty 3D array of SpaceTimeCube objects.
    Whenever a point value or contiguous space-time region of values is requested, identify
    which SpaceTimeCubes contain the requested coordinates, load them into the cache array
    by calling load_subcube for any that are not already present, request the values from
    these SpaceTimeCubes, combine them together if necessary, and return the resulting scalar
    value or array to the caller.
    """
    ndim         : cy.int
    size         : cy.ulonglong
    shape        : tuple[cy.int, cy.int, cy.int]
    subcube_shape: tuple[cy.int, cy.int, cy.int]
    cache_shape  : tuple[cy.int, cy.int, cy.int]
    cache        : ndarray
    load_subcube : object


    def __init__(self,
                 cube_shape   : tuple[int, int, int],
                 subcube_shape: tuple[int, int, int],
                 load_subcube : object) -> cy.void:
        """
        NOTE: The resolutions in cube_shape must be exact multiples of those in subcube_shape.
        """
        # Ensure that cube_shape and subcube_shape both contain 3 values or throw an error
        if len(cube_shape) != 3:
            raise ValueError("The cube_shape must contain exactly three values.")

        if len(subcube_shape) != 3:
            raise ValueError("The subcube_shape must contain exactly three values.")

        # Unpack the cube_shape values without type-checking
        cube_bands_: object = cube_shape[0]
        cube_rows_ : object = cube_shape[1]
        cube_cols_ : object = cube_shape[2]

        # Unpack the subcube_shape values without type-checking
        subcube_bands_: object = subcube_shape[0]
        subcube_rows_ : object = subcube_shape[1]
        subcube_cols_ : object = subcube_shape[2]

        # Ensure that cube_shape and subcube_shape only contain positive integers or throw an error
        if not(is_pos_int(cube_bands_) and is_pos_int(cube_rows_) and is_pos_int(cube_cols_)):
            raise ValueError("The cube_shape must contain only positive integers.")

        if not(is_pos_int(subcube_bands_) and is_pos_int(subcube_rows_) and is_pos_int(subcube_cols_)):
            raise ValueError("The subcube_shape must contain only positive integers.")

        # Cast the cube_shape values as primitive ints
        cube_bands: cy.int = cube_bands_
        cube_rows : cy.int = cube_rows_
        cube_cols : cy.int = cube_cols_

        # Cast the subcube_shape values as primitive ints
        subcube_bands: cy.int = subcube_bands_
        subcube_rows : cy.int = subcube_rows_
        subcube_cols : cy.int = subcube_cols_

        # Ensure that cube_shape is divided evenly by subcube_shape or throw an error
        cache_bands: cy.int = divide_evenly(cube_bands, subcube_bands)
        cache_rows : cy.int = divide_evenly(cube_rows, subcube_rows)
        cache_cols : cy.int = divide_evenly(cube_cols, subcube_cols)

        # Store the cube metadata, subcube_shape, cache_shape, cache, and load_subcube functions for later
        self.ndim          = 3
        self.size          = cube_bands * cube_rows * cube_cols
        self.shape         = (cube_bands, cube_rows, cube_cols)
        self.subcube_shape = (subcube_bands, subcube_rows, subcube_cols)
        self.cache_shape   = (cache_bands, cache_rows, cache_cols)
        self.cache         = np.empty(self.cache_shape, dtype=object)
        self.load_subcube  = load_subcube


    @cy.cfunc
    def __getOrLoadSubcube(self, cache_t: pyidx, cache_y: pyidx, cache_x: pyidx) -> SpaceTimeCube:
        """
        Return the SpaceTimeCube stored at self.cache[cache_t, cache_y, cache_x] if it
        has already been loaded. Otherwise, call self.load_subcube to load it, store
        it in self.cache, and return it.
        """
        subcube: SpaceTimeCube = cy.cast(SpaceTimeCube, self.cache[cache_t, cache_y, cache_x])
        if subcube:
            return subcube
        else:
            subcube = self.load_subcube((cache_t, cache_y, cache_x), self.subcube_shape)
            self.cache[cache_t, cache_y, cache_x] = subcube
            return subcube


    @cy.ccall
    @cy.exceptval(check=False)
    def get(self, t: pyidx, y: pyidx, x: pyidx) -> cy.float:
        """
        Return the scalar value at index (t,y,x) by translating these cube coordinates
        to cache and subcube coordinates, loading the matching subcube into the cache grid
        if not already present, and looking up the value within this subcube.

        NOTE: Indices may be negative provided that your load_subcube function can handle
              negative indices in its cache_index argument.
        """
        # Grab the subcube_shape tuple
        subcube_shape: tuple[cy.int, cy.int, cy.int] = self.subcube_shape

        # Unpack the subcube_shape values
        subcube_bands: cy.int = subcube_shape[0]
        subcube_rows : cy.int = subcube_shape[1]
        subcube_cols : cy.int = subcube_shape[2]

        # Calculate the cache index
        cache_t: pyidx = t // subcube_bands
        cache_y: pyidx = y // subcube_rows
        cache_x: pyidx = x // subcube_cols

        # Calculate the subcube index
        subcube_t: pyidx = t % subcube_bands
        subcube_y: pyidx = y % subcube_rows
        subcube_x: pyidx = x % subcube_cols

        # Fetch the subcube from the cache
        subcube: SpaceTimeCube = self.__getOrLoadSubcube(cache_t, cache_y, cache_x)

        # Look up the scalar value in the subcube at the subcube index
        return subcube.get(subcube_t, subcube_y, subcube_x)


    def getTimeSeries(self, t_range, y, x):
        """
        Return the 1D array given by the slice (t_range,y,x) by translating these cube
        coordinates to cache and subcube coordinates, loading the matching subcubes into
        the cache grid if not already present, looking up the array slices within each
        subcube, and merging them together into a single 1D array.

        NOTE: Indices may be negative provided that your load_subcube function can handle
              negative indices in its cache_index argument.
        NOTE: Range indices may include one or more None values and
              provide (inclusion, exclusion) semantics like Python array slice notation.
        """
        # Destructure the argument range
        (t_start, t_stop_exclusive) = to_positive_index_range(t_range, self.shape[0])
        t_stop = t_stop_exclusive - 1
        # Translate high-res coordinates to cache and subcube coordinates
        (subcube_bands, subcube_rows, subcube_cols) = self.subcube_shape
        (cache_t_start, subcube_t_start) = divmod(t_start, subcube_bands)
        (cache_t_stop,  subcube_t_stop)  = divmod(t_stop,  subcube_bands)
        (cache_y,       subcube_y)       = divmod(y,       subcube_rows)
        (cache_x,       subcube_x)       = divmod(x,       subcube_cols)
        # Load, expand, and combine subcubes
        return np.concatenate(
            [self.__getOrLoadSubcube(cache_t,
                                     cache_y,
                                     cache_x
                                    ).getTimeSeries(
                                        (subcube_t_start    if cache_t == cache_t_start else 0,
                                         subcube_t_stop + 1 if cache_t == cache_t_stop  else subcube_bands),
                                        subcube_y,
                                        subcube_x
                                    )
             for cache_t in range(cache_t_start, cache_t_stop + 1)]
        )


    def getSpatialPlane(self, t, y_range, x_range):
        """
        Return the 2D array given by the slice (t,y_range,x_range) by translating these
        cube coordinates to cache and subcube coordinates, loading the matching subcubes
        into the cache grid if not already present, looking up the array slices within each
        subcube, and merging them together into a single 2D array.

        NOTE: Indices may be negative provided that your load_subcube function can handle
              negative indices in its cache_index argument.
        NOTE: Range indices may include one or more None values and
              provide (inclusion, exclusion) semantics like Python array slice notation.
        """
        # Destructure the argument ranges
        (y_start, y_stop_exclusive) = to_positive_index_range(y_range, self.shape[1])
        (x_start, x_stop_exclusive) = to_positive_index_range(x_range, self.shape[2])
        y_stop = y_stop_exclusive - 1
        x_stop = x_stop_exclusive - 1
        # Translate high-res coordinates to cache and subcube coordinates
        (subcube_bands, subcube_rows, subcube_cols) = self.subcube_shape
        (cache_t,       subcube_t)       = divmod(t,       subcube_bands)
        (cache_y_start, subcube_y_start) = divmod(y_start, subcube_rows)
        (cache_y_stop,  subcube_y_stop)  = divmod(y_stop,  subcube_rows)
        (cache_x_start, subcube_x_start) = divmod(x_start, subcube_cols)
        (cache_x_stop,  subcube_x_stop)  = divmod(x_stop,  subcube_cols)
        # Load, expand, and combine subcubes
        return np.block(
            [[self.__getOrLoadSubcube(cache_t,
                                      cache_y,
                                      cache_x
                                      ).getSpatialPlane(
                                          subcube_t,
                                          (subcube_y_start    if cache_y == cache_y_start else 0,
                                           subcube_y_stop + 1 if cache_y == cache_y_stop  else subcube_rows),
                                          (subcube_x_start    if cache_x == cache_x_start else 0,
                                           subcube_x_stop + 1 if cache_x == cache_x_stop  else subcube_cols)
                                      )
              for cache_x in range(cache_x_start, cache_x_stop + 1)]
             for cache_y in range(cache_y_start, cache_y_stop + 1)]
        )


    def getSubcube(self, t_range, y_range, x_range):
        """
        Return the 3D array given by the slice (t_range,y_range,x_range) by translating
        these cube coordinates to cache and subcube coordinates, loading the matching
        subcubes into the cache grid if not already present, looking up the array slices
        within each subcube, and merging them together into a single 3D array.

        NOTE: Indices may be negative provided that your load_subcube function can handle
              negative indices in its cache_index argument.
        NOTE: Range indices may include one or more None values and
              provide (inclusion, exclusion) semantics like Python array slice notation.
        """
        # Destructure the argument ranges
        (t_start, t_stop_exclusive) = to_positive_index_range(t_range, self.shape[0])
        (y_start, y_stop_exclusive) = to_positive_index_range(y_range, self.shape[1])
        (x_start, x_stop_exclusive) = to_positive_index_range(x_range, self.shape[2])
        t_stop = t_stop_exclusive - 1
        y_stop = y_stop_exclusive - 1
        x_stop = x_stop_exclusive - 1
        # Translate high-res coordinates to cache and subcube coordinates
        (subcube_bands, subcube_rows, subcube_cols) = self.subcube_shape
        (cache_t_start, subcube_t_start) = divmod(t_start, subcube_bands)
        (cache_t_stop,  subcube_t_stop)  = divmod(t_stop,  subcube_bands)
        (cache_y_start, subcube_y_start) = divmod(y_start, subcube_rows)
        (cache_y_stop,  subcube_y_stop)  = divmod(y_stop,  subcube_rows)
        (cache_x_start, subcube_x_start) = divmod(x_start, subcube_cols)
        (cache_x_stop,  subcube_x_stop)  = divmod(x_stop,  subcube_cols)
        # Load, expand, and combine subcubes
        return np.block(
            [[[self.__getOrLoadSubcube(cache_t,
                                       cache_y,
                                       cache_x
                                       ).getSubcube(
                                           (subcube_t_start    if cache_t == cache_t_start else 0,
                                            subcube_t_stop + 1 if cache_t == cache_t_stop  else subcube_bands),
                                           (subcube_y_start    if cache_y == cache_y_start else 0,
                                            subcube_y_stop + 1 if cache_y == cache_y_stop  else subcube_rows),
                                           (subcube_x_start    if cache_x == cache_x_start else 0,
                                            subcube_x_stop + 1 if cache_x == cache_x_stop  else subcube_cols)
                                       )
               for cache_x in range(cache_x_start, cache_x_stop + 1)]
              for cache_y in range(cache_y_start, cache_y_stop + 1)]
             for cache_t in range(cache_t_start, cache_t_stop + 1)]
        )


    def getFullyRealizedCube(self, cache=False):
        raise ValueError("getFullyRealizedCube is not implemented for LazySpaceTimeCube.\n"
                         + "You probably don't want to do this anyway.")


    def releaseFullyRealizedCube(self):
        raise ValueError("releaseFullyRealizedCube is not implemented for LazySpaceTimeCube.\n"
                         + "You probably don't want to do this anyway.")
# lazy-space-time-cube-class ends here



# [[file:../../org/pyretechnics.org::units-conversion][units-conversion]]
import cython
import cython as cy
if cython.compiled:
    from cython.cimports.libc.math import pi, sqrt, sin, cos, tan, atan, atan2
    from cython.cimports.pyretechnics.cy_types import vec_xy
else:
    from math import pi, sqrt, sin, cos, tan, atan, atan2
    from pyretechnics.py_types import vec_xy


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def rad_to_deg(radians: cy.float) -> cy.float:
    """Convert radians to degrees."""
    return radians * 180.0 / pi


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def deg_to_rad(degrees: cy.float) -> cy.float:
    """Convert degrees to radians."""
    return degrees * pi / 180.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def deg_to_ratio(degrees: cy.float) -> cy.float:
    """Convert degrees to ratio."""
    return tan(deg_to_rad(degrees))


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def ratio_to_deg(ratio: cy.float) -> cy.float:
    """Convert ratio to degrees."""
    return rad_to_deg(atan(ratio))


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def F_to_K(degrees: cy.float) -> cy.float:
    """Convert fahrenheit to kelvin."""
    return (degrees + 459.67) * 0.5555555555555556


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def K_to_F(degrees: cy.float) -> cy.float:
    """Convert kelvin to fahrenheit."""
    return (degrees * 1.8) - 459.67


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def F_to_C(degrees: cy.float) -> cy.float:
    """Convert fahrenheit to celsius."""
    return (degrees - 32.0) * 0.5555555555555556


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def C_to_F(degrees: cy.float) -> cy.float:
    """Convert celsius to fahrenheit."""
    return (degrees * 1.8) + 32.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def ch_to_m(ch: cy.float) -> cy.float:
    """Convert chains to meters."""
    return ch * 20.1168


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def m_to_ch(m: cy.float) -> cy.float:
    """Convert meters to chains."""
    return m * 0.0497097


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def m_to_ft(m: cy.float) -> cy.float:
    """Convert meters to feet."""
    return m * 3.281


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def ft_to_m(ft: cy.float) -> cy.float:
    """Convert feet to meters."""
    return ft * 0.30478512648582745


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def mph_to_mps(mph: cy.float) -> cy.float:
    """Convert miles per hour to meters per second."""
    return mph * 0.44701818551254696


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def mps_to_mph(mps: cy.float) -> cy.float:
    """Convert meters per second to miles per hour."""
    return mps * 2.237045454545455


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def km_hr_to_mps(km_hr: cy.float) -> cy.float:
    """Convert kilometers per hour to meters per second."""
    return km_hr * 0.277764222883701


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def mps_to_km_hr(mps: cy.float) -> cy.float:
    """Convert meters per second to kilometers per hour."""
    return mps * 3.6001756800000004


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def mph_to_km_hr(mph: cy.float) -> cy.float:
    """Convert miles per hour to kilometers per hour."""
    return mph * 1.609344


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def km_hr_to_mph(km_hr: cy.float) -> cy.float:
    """Convert kilometers per hour to miles per hour."""
    return km_hr * 0.621371192237334


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def m_min_to_km_hr(m_min: cy.float) -> cy.float:
    """Convert meters per minute to kilometers per hour."""
    return m_min * 0.06


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def km_hr_to_m_min(km_hr: cy.float) -> cy.float:
    """Convert kilometers per hour to meters per minute."""
    return km_hr / 0.06


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def m_min_to_mph(m_min: cy.float) -> cy.float:
    """Convert meters per minute to miles per hour."""
    return m_min * 0.0372840909091


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def mph_to_m_min(mph: cy.float) -> cy.float:
    """Convert miles per hour to meters per minute."""
    return mph * 26.8210911307


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def mps_to_fpm(mps: cy.float) -> cy.float:
    """Convert meters per second to feet per minute."""
    return mps * 196.86


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def fpm_to_mps(fpm: cy.float) -> cy.float:
    """Convert feet per minute to meters per second."""
    return fpm / 196.86


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def mph_to_fpm(mph: cy.float) -> cy.float:
    """Convert miles per hour to feet per minute."""
    return mph * 88.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def fpm_to_mph(fpm: cy.float) -> cy.float:
    """Convert feet per minute to miles per hour."""
    return fpm / 88.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def Btu_ft_s_to_kW_m(Btu_ft_s: cy.float) -> cy.float:
    """Convert BTU per feet per second to kilowatt per meter."""
    return Btu_ft_s * 3.46165186


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def kW_m_to_Btu_ft_s(kW_m: cy.float) -> cy.float:
    """Convert kilowatt per meter to BTU per feet per second."""
    return kW_m * 0.28887942532730604


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def Btu_lb_to_kJ_kg(Btu_lb: cy.float) -> cy.float:
    """Convert BTU per lb to kilojoule per kilogram."""
    return Btu_lb * 2.3259999996185


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def kJ_kg_to_Btu_lb(kJ_kg: cy.float) -> cy.float:
    """Convert kilojoule per kilogram to BTU per lb."""
    return kJ_kg / 2.3259999996185


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def kg_m3_to_lb_ft3(kg_m3: cy.float) -> cy.float:
    """Convert kilogram per cubic meter to pound per cubic foot."""
    return kg_m3 * 0.0624


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def lb_ft3_to_kg_m3(lb_ft3: cy.float) -> cy.float:
    """Convert pound per cubic foot to kilogram per cubic meter."""
    return lb_ft3 * 16.025641025641026


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def percent_to_dec(percent: cy.float) -> cy.float:
    """Convert percent to decimal."""
    return percent * 0.01


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def dec_to_percent(decimal: cy.float) -> cy.float:
    """Convert decimal to percent."""
    return decimal * 100.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def sec_to_min(seconds: cy.float) -> cy.float:
    """Convert seconds to minutes."""
    return seconds * 0.016666666666666666


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def min_to_sec(minutes: cy.float) -> cy.float:
    """Convert minutes to seconds."""
    return minutes * 60.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def ms_to_min(milliseconds: cy.float) -> cy.float:
    """Convert milliseconds to minutes."""
    return milliseconds * 0.000016667


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def min_to_ms(minutes: cy.float) -> cy.float:
    """Convert minutes to milliseconds."""
    return minutes * 60000.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def hour_to_min(hours: cy.float) -> cy.float:
    """Converts hours to minutes."""
    return hours * 60.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def min_to_hour(minutes: cy.float) -> cy.float:
    """Converts minutes to hours."""
    return minutes / 60.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def day_to_min(days: cy.float) -> cy.float:
    """Convert days to minutes."""
    return days * 1440.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def min_to_day(minutes: cy.float) -> cy.float:
    """Convert minutes to days."""
    return minutes / 1440.0


# TODO Return vec_rt
@cy.ccall
@cy.exceptval(check=False)
def cartesian_to_polar(x: cy.float, y: cy.float) -> vec_xy:
    """Convert cartesian coordinates (x, y) to polar coordinates (r, theta)."""
    r        : cy.float = sqrt(x * x + y * y)
    theta_rad: cy.float = atan2(y, x)
    theta    : cy.float = (rad_to_deg(theta_rad) + 360.0) % 360.0
    return (r, theta)


@cy.ccall
@cy.exceptval(check=False)
def polar_to_cartesian(r: cy.float, theta: cy.float) -> vec_xy:
    """Convert polar coordinates (r, theta) to cartesian coordinates (x, y)."""
    theta_rad: cy.float = deg_to_rad(theta)
    x        : cy.float = r * cos(theta_rad)
    y        : cy.float = r * sin(theta_rad)
    return (x, y)


# TODO Return vec_ra
@cy.ccall
@cy.exceptval(check=False)
def cartesian_to_azimuthal(x: cy.float, y: cy.float) -> vec_xy:
    """Convert cartesian coordinates (x, y) to azimuthal coordinates (r, azimuth)."""
    r          : cy.float = sqrt(x * x + y * y)
    azimuth_rad: cy.float = atan2(x, y)
    azimuth    : cy.float = (rad_to_deg(azimuth_rad) + 360.0) % 360.0
    return (r, azimuth)


@cy.ccall
@cy.exceptval(check=False)
def azimuthal_to_cartesian(r: cy.float, azimuth: cy.float) -> vec_xy:
    """Convert azimuthal coordinates (r, azimuth) to cartesian coordinates (x, y)."""
    azimuth_rad: cy.float = deg_to_rad(azimuth)
    x          : cy.float = r * sin(azimuth_rad)
    y          : cy.float = r * cos(azimuth_rad)
    return (x, y)


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def opposite_direction(theta: cy.float) -> cy.float:
    """Convert theta to theta + 180 degrees."""
    return (theta + 180.0) % 360.0


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def wind_speed_10m_to_wind_speed_20ft(wind_speed_10m: cy.float) -> cy.float:
    """Convert wind speed at 10m to wind speed at 20ft."""
    return wind_speed_10m / 1.15


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def wind_speed_20ft_to_wind_speed_10m(wind_speed_20ft: cy.float) -> cy.float:
    """Convert wind speed at 20ft to wind speed at 10m."""
    return wind_speed_20ft * 1.15
# units-conversion ends here



import os
import numpy
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext
from Cython.Build import cythonize

set_gcc_args   = (os.getenv("PYR_SET_GCC_ARGS")   == "1")
profile_cython = (os.getenv("PYR_PROFILE_CYTHON") == "1")

class custom_build_ext(build_ext):
    def build_extensions(self):
        if set_gcc_args:
            # This removes the "default" compiler flags that would
            # otherwise get passed on to to the compiler, i.e.,
            # sysconfig.get_config_var("CFLAGS").
            self.compiler.set_executable("compiler_so", "gcc")
        build_ext.build_extensions(self)

    def build_extension(self, extension):
        if set_gcc_args:
            extension.extra_compile_args = [
                  # Warnings
                  "-Wall",
                  "-Wno-maybe-uninitialized",
                  "-Wno-unused-result",
                  "-Wno-unused-function",
                  "-Wsign-compare",
                  # Optimization
                  "-O3",
                  "-fno-semantic-interposition",
                  # Debugging
                  "-g",
                  "-DNDEBUG",
                  "-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION",
                  # Code Generation
                  "-fwrapv",
                  "-fPIC",
            ]
        build_ext.build_extension(self, extension)

extensions = [
    Extension("*", ["src/pyretechnics/*.py"], include_dirs=[numpy.get_include()])
]

setup(name="pyretechnics",
      cmdclass={"build_ext": custom_build_ext},
      ext_modules=cythonize(extensions,
                            exclude="src/pyretechnics/py_types.py",
                            annotate=True,
                            compiler_directives={
                                "language_level": "3",
                                "profile": profile_cython,
                                "initializedcheck": False,
                                "cdivision": True,
                                "wraparound": False,
                                "boundscheck": False,
                            }))



# [[file:../../org/pyretechnics.org::vector-utilities][vector-utilities]]
import cython
import cython as cy
if cython.compiled:
    from cython.cimports.libc.math import sqrt, sin, cos
    from cython.cimports.pyretechnics.cy_types import pyidx, vec_xy, vec_xyz
    import cython.cimports.pyretechnics.conversion as conv
else:
    from math import sqrt, sin, cos
    from pyretechnics.py_types import pyidx, vec_xy, vec_xyz
    import pyretechnics.conversion as conv


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def dot_2d(vector1: vec_xy, vector2: vec_xy) -> cy.float:
    return vector1[0] * vector2[0] + vector1[1] * vector2[1]


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def dot_3d(vector1: vec_xyz, vector2: vec_xyz) -> cy.float:
    return vector1[0] * vector2[0] + vector1[1] * vector2[1] + vector1[2] * vector2[2]


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def scale_2d(scalar: cy.float, vector: vec_xy) -> vec_xy:
    return (scalar * vector[0], scalar * vector[1])


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def scale_3d(scalar: cy.float, vector: vec_xyz) -> vec_xyz:
    return (scalar * vector[0], scalar * vector[1], scalar * vector[2])


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def add_2d(vector1: vec_xy, vector2: vec_xy) -> vec_xy:
    return (vector1[0] + vector2[0], vector1[1] + vector2[1])


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def add_3d(vector1: vec_xyz, vector2: vec_xyz) -> vec_xyz:
    return (vector1[0] + vector2[0], vector1[1] + vector2[1], vector1[2] + vector2[2])


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def vector_magnitude_2d(vector: vec_xy) -> cy.float:
    return sqrt(dot_2d(vector, vector))


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def vector_magnitude_3d(vector: vec_xyz) -> cy.float:
    return sqrt(dot_3d(vector, vector))


@cy.ccall
@cy.exceptval(check=False)
def as_unit_vector_2d(vector: vec_xy) -> vec_xy:
    magnitude: cy.float = vector_magnitude_2d(vector)
    if magnitude == 0.0:
        return vector
    else:
        ux: cy.float = vector[0] / magnitude
        uy: cy.float = vector[1] / magnitude
        return (ux, uy)


@cy.ccall
@cy.exceptval(check=False)
def as_unit_vector_3d(vector: vec_xyz) -> vec_xyz:
    magnitude: cy.float = vector_magnitude_3d(vector)
    if magnitude == 0.0:
        return vector
    else:
        ux: cy.float = vector[0] / magnitude
        uy: cy.float = vector[1] / magnitude
        uz: cy.float = vector[2] / magnitude
        return (ux, uy, uz)


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def to_slope_plane(vector_2d: vec_xy, elevation_gradient: vec_xy) -> vec_xyz:
    return (
        vector_2d[0],
        vector_2d[1],
        dot_2d(vector_2d, elevation_gradient)
    )


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def to_horizontal_plane(vector_3d: vec_xyz) -> vec_xy:
    return (vector_3d[0], vector_3d[1])


@cy.ccall
@cy.exceptval(check=False)
def spread_direction_vector_to_angle(vector_3d: vec_xyz) -> cy.float:
    x        : cy.float = vector_3d[0]
    y        : cy.float = vector_3d[1]
    az_coords: vec_xy   = conv.cartesian_to_azimuthal(x, y)
    azimuth  : cy.float = az_coords[1]
    return azimuth


@cy.ccall
@cy.exceptval(check=False)
def get_slope_normal_vector(elevation_gradient: vec_xy) -> vec_xyz:
    (dz_dx, dz_dy)               = elevation_gradient
    slope_normal_vector: vec_xyz = (-dz_dx, -dz_dy, 1.0)
    return as_unit_vector_3d(slope_normal_vector)


@cy.ccall
@cy.exceptval(check=False)
def cross_3d(vector1: vec_xyz, vector2: vec_xyz) -> vec_xyz:
    (a, b, c) = vector1
    (d, e, f) = vector2
    return (
        b * f - e * c,
        -a * f + d * c,
        a * e - d * b,
    )


@cy.ccall
@cy.exceptval(check=False)
def rotate_on_sloped_plane(vector: vec_xyz, theta: cy.float, slope: cy.float, aspect: cy.float) -> vec_xyz:
    """
    Rotate a 3D vector <x,y,z> theta degrees clockwise on the plane defined by the slope and aspect.
    """
    # Calculate the slope normal vector from the slope and aspect
    elevation_gradient : vec_xy  = conv.azimuthal_to_cartesian(slope, conv.opposite_direction(aspect))
    slope_normal_vector: vec_xyz = get_slope_normal_vector(elevation_gradient)
    # Calculate sine and cosine of theta
    theta_rad: cy.float = conv.deg_to_rad(theta)
    cos_theta: cy.float = cos(theta_rad)
    sin_theta: cy.float = sin(theta_rad)
    # Rotate theta degrees clockwise around the slope_normal_vector
    vector_i: vec_xyz = (
        cos_theta * vector[0],
        cos_theta * vector[1],
        cos_theta * vector[2],
    )
    vector_j: vec_xyz = (
        sin_theta * vector[0],
        sin_theta * vector[1],
        sin_theta * vector[2],
    )
    vector_k: vec_xyz = cross_3d(vector_j, slope_normal_vector)
    return (
        vector_i[0] + vector_k[0],
        vector_i[1] + vector_k[1],
        vector_i[2] + vector_k[2],
    )
# vector-utilities ends here
